// GameBoard.cpp : implementation file
//

#include "GameCommon.h"
#include "GameBoard.h"
#include "GameCardStack.h"
#include "GamePlayer.h"
#include "GameModel.h"
#include "GameBoardRenderer.h"

#include "logging.h" //LogDebugString

/////////////////////////////////////////////////////////////////////////////
// CGameBoard

CGameBoard::CGameBoard(CGameModel& aModel, EBoardLayout aLayout) : 
	iGameModel(aModel), iBoardLayout(aLayout), iCardStackMove(0), iCardMove(0),
	iFaceDownAtStartMove(false), iNumberOfInStacks(aModel.NumberOfInStacks())
{
}

CGameBoard::~CGameBoard()
{
}

#ifdef OS_SYMBIAN
void CGameBoard::ConstructL()
{
	iGameModel.AddObserver(*this);
}
#else // not OS_SYMBIAN
bool CGameBoard::InitResources()
{
	iGameModel.AddObserver(*this);
	return true;
}
#endif

void CGameBoard::NewGame(int /*aNumberOfPlayers*/)
{
	CGameCardStack& deck = static_cast<CGameCardStack&>(iGameModel.CardDeck());	

	iNumberOfInStacks = iGameModel.NumberOfInStacks();

	for (int i = 0; i < iNumberOfInStacks; ++i)
	{
		deck.MoveCard(iGameModel.GameState(), deck.Card(51-i), 
			static_cast<CGameCardStack&>(iGameModel.CardInStack(i)));
	}

	for (int i = 4; i < 52; ++i)
	{
		deck.MoveCard(iGameModel.GameState(), deck.Card(51-i), static_cast<CGameCardStack&> 
			(iGameModel.PlayerInHand(i % 2)));
	}
	
	SetSpacialInformation(iX, iY, iWidth, iHeight);
}

bool CGameBoard::SerializeIn(const unsigned char*& aBufferPtr, int& aBufferBytesLeft)
{
	if (aBufferBytesLeft < 1)
		return false;
	
	iNumberOfInStacks = *(aBufferPtr++);
	--aBufferBytesLeft;

	return true;
}

bool CGameBoard::SerializeOut(unsigned char*& aBufferPtr, int& aBufferBytesLeft) const
{
	if (aBufferBytesLeft < 1)
		return false;
	
	*(aBufferPtr++) = iNumberOfInStacks;
	--aBufferBytesLeft;

	return true;
}

void CGameBoard::SpacialInformation(int& aX, int& aY, int& aWidth, int& aHeight) const
{
	SSpacialInformation::SpacialInformation(aX, aY, aWidth, aHeight);
} 

void CGameBoard::SetSpacialInformation(int aX, int aY, int aWidth, int aHeight)
{
	SSpacialInformation::SetSpacialInformation(aX, aY, aWidth, aHeight);

	switch (iBoardLayout)
	{
	case ELayoutTower:
	default:
		{
			int columnWidth = iWidth / iNumberOfInStacks;
			int cardWidth = iCardImageSize.iWidth;
			int cardHeight = iCardImageSize.iHeight;
			int rowHeight = (cardHeight * 4) / 3;
			if (rowHeight * 6 > iHeight)
			{
				// minimum of 6 "rows"
				rowHeight = iHeight / 6;
			}
			int cardXOffset = (columnWidth - cardWidth) / 2;
			int cardYOffset = (rowHeight - cardHeight) / 2;

			for (int i = 0; i < 52; ++i)
			{
				static_cast<SSpacialInformation&>(iGameModel.CardDeck().Card(i)).
					SetSpacialInformation(0, 0, cardWidth, cardHeight);
			}

			for (int i = 0; i < iNumberOfInStacks; ++i)
			{
				if (i < 4)
					static_cast<SSpacialInformation&>(iGameModel.CardOutStack(i)).
						SetSpacialInformation(cardXOffset + (i * columnWidth), 
						cardYOffset, 
						cardWidth, 
						cardHeight);
				static_cast<SSpacialInformation&>(iGameModel.CardInStack(i)).
					SetSpacialInformation(cardXOffset + (i * columnWidth), 
					cardYOffset + rowHeight, 
					cardWidth, 
					iHeight - (2 * (rowHeight + cardYOffset)));
			}
			static_cast<SSpacialInformation&>(iGameModel.PlayerInHand(0)).
				SetSpacialInformation(cardXOffset, 
				(iHeight - rowHeight) + cardYOffset, 
				cardWidth, 
				cardHeight);
			static_cast<SSpacialInformation&>(iGameModel.PlayerOutHand(0)).
				SetSpacialInformation(cardXOffset + columnWidth, 
				(iHeight - rowHeight) + cardYOffset, 
				cardWidth, 
				cardHeight);
			static_cast<SSpacialInformation&>(iGameModel.PlayerOutHand(1)).
				SetSpacialInformation(cardXOffset + (columnWidth * 2), 
				(iHeight - rowHeight) + cardYOffset, 
				cardWidth, 
				cardHeight);
			static_cast<SSpacialInformation&>(iGameModel.PlayerInHand(1)).
				SetSpacialInformation(cardXOffset + (columnWidth * 3), 
				(iHeight - rowHeight) + cardYOffset, 
				cardWidth, 
				cardHeight);
		}
		break;
	}
}

bool CGameBoard::StartMove(int aX, int aY, bool aOutPass)
{
	CGameCardStack* srcStack = 0;
	CGameCard* srcCard = 0;
	if (!CardAtLocation(aX, aY, srcStack, srcCard))
	{
		return false;
	}
	iStartMoveOffsetX = aX - srcCard->iX;
	iStartMoveOffsetY = aY - srcCard->iY;
	return StartMove(*srcStack, *srcCard, aOutPass);
}

bool CGameBoard::StartMove(CGameCardStack& aSrcStack, CGameCard& aSrcCard, bool aOutPass)
{
	EGameState state = iGameModel.GameState();

	switch(state)
	{
	case EGameStatePlayer1ToMove:
		break;
	case EGameStatePlayer2ToMove:
		break;
	case EGameStatePlayer1MadeMove:
		state = EGameStatePlayer1ToMove;
		break;
	case EGameStatePlayer2MadeMove:
		state = EGameStatePlayer2ToMove;
		break;
	case EGameStatePlayer1MakingMove:
		state = EGameStatePlayer1ToMove;
		break;
	case EGameStatePlayer2MakingMove:
		state = EGameStatePlayer2ToMove;
		break;
	default:
		return false;
	}

	iGameModel.SetGameState(state);

	iCardStackMove = 0;
	iCardMove = 0;

	if (!(state & EGameStateToMove))
	{
		return false;
	}

	iCardStackMove = &aSrcStack;
	iCardMove = &aSrcCard;

	iFaceDownAtStartMove = iCardMove->FaceDown();

	CGameCardStack::EType type = iCardStackMove->Type();

	if (type == CGameCardStack::EOutStack || 
		(type == CGameCardStack::EPlayer2InHand && state & EGameStatePlayer1) ||
		(type == CGameCardStack::EPlayer1InHand && state & EGameStatePlayer2))
	{
		return false;
	}

	switch(state)
	{
	case EGameStatePlayer1ToMove:
		iGameModel.SetGameState(EGameStatePlayer1MakingMove);
		break;
	case EGameStatePlayer2ToMove:
		iGameModel.SetGameState(EGameStatePlayer2MakingMove);
		break;
	default:
		break;
	}

	iCardMove->SetFaceDown(false);

	return true;
}

bool CGameBoard::EndMove(int /*aX*/, int /*aY*/, bool aOutPass)
{
	CGameCardStack* targetStack = 0;
	CGameCard* targetCard = 0;
	if (!CardStackAtLocation(iCardMove->iX + (iCardMove->iWidth / 2), 
		iCardMove->iY + (iCardMove->iHeight / 2), targetStack, targetCard))
	{
		return false;
	}
	return EndMove(*targetStack, aOutPass);
}

bool CGameBoard::EndMove(CGameCardStack& aTargetStack, bool aOutPass)
{
	EGameState state = iGameModel.GameState();

	switch(state)
	{
	case EGameStatePlayer1MakingMove:
		break;
	case EGameStatePlayer2MakingMove:
		break;
	default:
		return false;
	}

	CGameCardStack* targetStack = 0;

	bool invalidOutPass(false);

	if (aOutPass)
	{
		int outStackIndex = 0;
		for (; outStackIndex < 4; ++outStackIndex)
		{
			CGameCardStack* stack = &iGameModel.CardOutStack(outStackIndex);
			if (iCardStackMove->CanMoveCard(state, *iCardMove, *stack))
			{
				targetStack = stack;
				break;
			}
		}
		if (outStackIndex == 4)
		{
			targetStack = (state == EGameStatePlayer1MakingMove ?
				&iGameModel.PlayerOutHand(0) : &iGameModel.PlayerOutHand(1));
			if (!iCardStackMove->CanMoveCard(state, *iCardMove, *targetStack))
				invalidOutPass = true;
		}
	}
	else
	{
		targetStack = &aTargetStack;
	}

	if (iCardStackMove->CanMoveCard(state, *iCardMove, *targetStack))
	{
		if (iGameModel.CurrentPlayer().ComputerPlayer())
			AnimateCardMove(*targetStack);

		iCardStackMove->MoveCard(state, *iCardMove, *targetStack);

		switch(state)
		{
		case EGameStatePlayer1MakingMove:
			switch(targetStack->Type())
			{
			case CGameCardStack::EPlayer1OutHand:
				iGameModel.SetGameState(EGameStatePlayer1MadeMove);
				break;
			default:
				iGameModel.SetGameState(EGameStatePlayer1ToMove);
				break;
			}
			break;
		case EGameStatePlayer2MakingMove:
			switch(targetStack->Type())
			{
			case CGameCardStack::EPlayer2OutHand:
				iGameModel.SetGameState(EGameStatePlayer2MadeMove);
				break;
			default:
				iGameModel.SetGameState(EGameStatePlayer2ToMove);
				break;
			}
			break;
		default:
			break;
		}
		return true;
	}
	else
	{
		if (!aOutPass || invalidOutPass)
		{
			iGameModel.SetGameState(state & EGameStatePlayer1 ? 
				EGameStatePlayer1ToMove : EGameStatePlayer2ToMove);
		}
		else
		{
			iGameModel.SetGameState(state & EGameStatePlayer1 ? 
				EGameStatePlayer2ToMove : EGameStatePlayer1ToMove);
		}
		return false;
	}
}

bool CGameBoard::CardStackAtLocation(int aX, int aY, CGameCardStack*& aGameCardStack, CGameCard*& aGameCard) const
{
	aGameCardStack = 0;
	aGameCard = 0;
	for (int i = 0; i < iGameModel.StackCount(); ++i)
	{
		CGameCardStack& stack = iGameModel.CardStack(i);
		if (stack.Type() == CGameCardStack::EDeck)
			continue;
		if (stack.PointIntersects(aX, aY))
		{
			aGameCardStack = &stack;
			for (int j = stack.CardCount() - 1; j >= 0; --j)
			{
				CGameCard& card  = stack.Card(j);
				if (card.PointIntersects(aX, aY))
				{
					aGameCard = &card;
					return true;
				}
			}
			return true;
		}
	}
	return false;
}

bool CGameBoard::CardAtLocation(int aX, int aY, CGameCardStack*& aGameCardStack, CGameCard*& aGameCard) const
{
	return CardStackAtLocation(aX, aY, aGameCardStack, aGameCard) && aGameCard != 0;
}

void CGameBoard::SelectionBoundingRect(int& aX1, int& aY1, int& aX2, int& aY2, bool aInit)
{
	for (int i = iCardStackMove->CardIndex(*iCardMove); i >= 0 && i < iCardStackMove->CardCount(); ++i)
	{
		int x1, y1, x2, y2, w, h;
		static_cast<ISpacialInformation&>(iCardStackMove->Card(i)).SpacialInformation(x1, y1, w, h);
		x2 = x1 + w;
		y2 = y1 + h;
		if (aInit)
		{
			aX1 = x1; aY1 = y1; aX2 = x2; aY2 = y2;
			aInit = false;
		}
		else
		{
			if (aX1 > x1)
				aX1 = x1;
			if (aY1 > y1)
				aY1 = y1;
			if (aX2 < x2)
				aX2 = x2;
			if (aY2 < y2)
				aY2 = y2;
		}
	}
}

void CGameBoard::MoveSelection(int aX, int aY)
{
	int oldXPos, oldYPos, oldWidth, oldHeight;
	static_cast<ISpacialInformation*>(iCardMove)->SpacialInformation(oldXPos, oldYPos, oldWidth, oldHeight);
	int offsetX = aX - oldXPos - iStartMoveOffsetX;
	int offsetY = aY - oldYPos - iStartMoveOffsetY;
	for (int i = iCardStackMove->CardIndex(*iCardMove); i >= 0 && i < iCardStackMove->CardCount(); ++i)
	{
		static_cast<ISpacialInformation&>(iCardStackMove->Card(i)).SpacialInformation(oldXPos, oldYPos, oldWidth, oldHeight);
		static_cast<ISpacialInformation&>(iCardStackMove->Card(i)).SetSpacialInformation(oldXPos+offsetX, oldYPos+offsetY, oldWidth, oldHeight);
	}
}

void CGameBoard::AnimateCardMove(CGameCardStack& aTarget)
{
	int x1, y1, w1, h1;
	static_cast<ISpacialInformation*>(iCardMove)->SpacialInformation(x1, y1, w1, h1);

	int x2, y2, w2, h2;
	static_cast<ISpacialInformation&>(aTarget).SpacialInformation(x2, y2, w2, h2);
	
	iStartMoveOffsetX = 0;
	iStartMoveOffsetY = 0;

	{
		/* Bresenham line drawing algorithm - from Dr Dobb's May 89 Graphics Programming*/
	 int  w, h, d, dxd, dyd, dxn, dyn, dinc, ndinc, p;
	 register x, y;
 
	   /* Set up */
	   x = x1; y = y1;                          /* start of line */
	   w = x2 - x1;                      /* width domain of line */
	   h = y2 - y1;                     /* height domain of line */
 
	   /* Determine drawing direction */
	   if (w < 0) {                     /* drawing right to left */
		 w = -w;                               /* absolute width */
		 dxd = -1;                    /* x increment is negative */
	   } else                           /* drawing left to right */
		 dxd = 1;                       /* so x incr is positive */
	   if (h < 0) {                     /* drawing bottom to top */
		 h = -h;                       /* so get absolute height */
		 dyd = -1;                         /* y incr is negative */
	   } else                           /* drawing top to bottom */
		 dyd = 1;                       /* so y incr is positive */
 
	   /* Determine major axis of motion */
	   if (w < h) {                           /* major axis is Y */
		 p = h, h = w, w = p;       /* exchange height and width */
		 dxn = 0;
		 dyn = dyd;
	   } else {                               /* major axis is X */
		 dxn = dxd;
		 dyn = 0;
	   }
 
	   /* Set control variables */
	   ndinc = h * 2;                  /* Non-diagonal increment */
	   d = ndinc - w;                /* pixel selection variable */
	   dinc = d - w;                       /* Diagonal increment */
 
	int sx1,sy1,sx2,sy2;
	bool init = false;
 	
	   /* Loop to draw the line */
	   for (p = 0; p <= w; p++) {
			if (!init)
			{
				SelectionBoundingRect(sx1,sy1,sx2, sy2, true);
				init = true;
			}
			MoveSelection (x, y);
		 SelectionBoundingRect(sx1, sy1, sx2, sy2, false);
		 if (p % iGameModel.CardSpeed() == 0)
			if (iGameBoardRenderer)
			{
				SSpacialInformation si(sx1,sy1,(sx2 - sx1) + 1, (sy2 - sy1) + 1);
				iGameBoardRenderer->RenderGameBoard(si);
				init = false;
			}
		 if (d < 0) {                     /* step non-diagonally */
		   x += dxn;
		   y += dyn;
		   d += ndinc;
		 } else {                             /* step diagonally */
		   x += dxd;
		   y += dyd;
		   d += dinc;
		 }
	   }
	 }
}

void CGameBoard::OnGameStateChanged(CGameModel& /*aModel*/, EGameState aOldState, EGameState aNewState)
{
	// first check for game over, TODO: stalemate detection
	switch(aNewState)
	{
	case EGameStatePlayer1MadeMove:
	case EGameStatePlayer1ToMove:
	case EGameStatePlayer2MadeMove:
	case EGameStatePlayer2ToMove:
		if (iGameModel.PlayerInHand(0).CardCount() == 0 &&
			iGameModel.PlayerOutHand(0).CardCount() == 0)
		{
			iGameModel.SetGameState(EGameStatePlayer1Won);
			return;
		}
		else if (iGameModel.PlayerInHand(1).CardCount() == 0 &&
			iGameModel.PlayerOutHand(1).CardCount() == 0)
		{
			iGameModel.SetGameState(EGameStatePlayer2Won);
			return;
		}
		break;
	default:
		break;
	}

	// next flip a full player out stack to a full player in stack after opponent has moved
	switch (aNewState)
	{
	case EGameStatePlayer1ToMove:
		if (iGameModel.PlayerInHand(0).CardCount() == 0)
		{
			int index = iGameModel.PlayerOutHand(0).CardCount() - 1;
			while (index >= 0)
			{
				CGameCard& card = iGameModel.PlayerOutHand(0).Card(index);
				iGameModel.PlayerInHand(0).AddCard(card);
				iGameModel.PlayerOutHand(0).RemoveCard(card);
				--index;
			}
		}
		break;
	case EGameStatePlayer1MadeMove:
		iGameModel.SetGameState(EGameStatePlayer2ToMove);
		break;
	case EGameStatePlayer2ToMove:
		if (iGameModel.PlayerInHand(1).CardCount() == 0)
		{
			int index = iGameModel.PlayerOutHand(1).CardCount() - 1;
			while (index >= 0)
			{
				CGameCard& card = iGameModel.PlayerOutHand(1).Card(index);
				iGameModel.PlayerInHand(1).AddCard(card);
				iGameModel.PlayerOutHand(1).RemoveCard(card);
				--index;
			}
		}
		break;
	case EGameStatePlayer2MadeMove:
		iGameModel.SetGameState(EGameStatePlayer1ToMove);
		break;
	default:
		break;
	}
}
