// GameCardStack.cpp : implementation file
//

#include "options.h"

#if !defined(OS_SYMBIAN)
#include <stdlib.h>
#include <time.h>
#else  // OS_SYMBIAN
#include <e32math.h>
#endif

#include "GameCardStack.h"

#include "GameModel.h"
#include "GameCard.h"

/////////////////////////////////////////////////////////////////////////////
// CGameCardStack

CGameCardStack::CGameCardStack(CGameModel& aModel, EType aType) : 
	iModel(aModel), iType(aType), iCardCount(0)
{
	for (int i = 0; i < 52; ++i)
		iCards[i] = 0;
}

CGameCardStack::~CGameCardStack()
{
	if (iType != EDeck)
		return; // nothing to destroy

	for (int i = 0; i < 52; ++i)
		delete iCards[i];
}

#ifdef OS_SYMBIAN
void CGameCardStack::ConstructL()
#else // not OS_SYMBIAN
bool CGameCardStack::InitResources()
#endif
{
	if (iType != EDeck)
		#ifdef OS_SYMBIAN
		return; // nothing to create;
		#else // not OS_SYMBIAN
		return true; // nothing to create;
		#endif

	CGameCard::ECardSuit suits[4] = 
	{ 		
		CGameCard::EClubs, CGameCard::EDiamonds, CGameCard::EHearts, CGameCard::ESpades
	};
	CGameCard::ECardValue values[13] = 
	{
		CGameCard::EAce, CGameCard::ETwo, CGameCard::EThree, CGameCard::EFour, 
		CGameCard::EFive, CGameCard::ESix, CGameCard::ESeven, CGameCard::EEight, 
		CGameCard::ENine, CGameCard::ETen, CGameCard::EJack, CGameCard::EQueen, 
		CGameCard::EKing
	};

	for (int i = 0; i < ENumberOfCards; ++i)
	{
#ifdef OS_SYMBIAN
		AddCard(*(new (ELeave) CGameCard(values[i % 13], suits[i / 13])));
#else // not OS_SYMBIAN
		AddCard(*(new CGameCard(values[i % 13], suits[i / 13])));
#endif
	}
	
	for (int i = 0; i < ENumberOfCards; ++i)
	{
#ifdef OS_SYMBIAN
		static_cast<IGameObject*>(iCards[i])->ConstructL();
#else // not OS_SYMBIAN
		static_cast<IGameObject*>(iCards[i])->InitResources();
#endif
	}

#ifndef OS_SYMBIAN
	return true;
#endif
}

void CGameCardStack::NewGame(int aNumberOfPlayers)
{
	int i;

	switch(iType)
	{
	case EDeck:
		for (i = 0; i < ENumberOfCards; ++i)
		{
			static_cast<IGameObject*>(iCards[i])->NewGame(aNumberOfPlayers);
		}
		iCardCount = 52;
		Shuffle();
		break;
	default:
		for (i = 0; i < ENumberOfCards; ++i)
		{
			iCards[i] = 0;
		}
		iCardCount = 0;
		break;
	}
}

bool CGameCardStack::SerializeIn(const unsigned char*& aBufferPtr, int& aBufferBytesLeft)
{
	if (aBufferBytesLeft < 2)
		return false;
 
	iType = static_cast<EType>(*aBufferPtr++);
	--aBufferBytesLeft;
	iCardCount = *aBufferPtr++;
	--aBufferBytesLeft;

	for (int i = 0; i < iCardCount; ++i)
	{
		if (!static_cast<ISerializable*>(iCards[i])->SerializeIn(aBufferPtr, aBufferBytesLeft))
			return false;
	}

	return true;
}

bool CGameCardStack::SerializeOut(unsigned char*& aBufferPtr, int& aBufferBytesLeft) const
{
	if (aBufferBytesLeft < 2)
		return false;

	*aBufferPtr++ = static_cast<unsigned char>(iType);
	--aBufferBytesLeft;
	*aBufferPtr++ = static_cast<unsigned char>(iCardCount);
	--aBufferBytesLeft;

	for (int i = 0; i < iCardCount; ++i)
	{
		if (!static_cast<ISerializable*>(iCards[i])->SerializeOut(aBufferPtr, aBufferBytesLeft))
			return false;
	}

	return true;
}

void CGameCardStack::Shuffle()
{
	for (int i = 0; i < 52; ++i)
	{
		iCards[i]->SetValue(static_cast<CGameCard::ECardValue>(i % 13));
		iCards[i]->SetSuit(static_cast<CGameCard::ECardSuit>(i / 13));
	}
	
	#if !defined(OS_SYMBIAN)
	int seed = time(NULL);
	srand(seed);
	int goes =  rand() % 50 + 5;
	#else // OS_SYMBIAN
	TInt64 seed(User::TickCount());
	int goes =  Math::Rand(seed) % 50 + 5;
	#endif

	for (int i = 0; i < goes; ++i)
	{
		#if !defined(OS_SYMBIAN)
		int sticky = rand() % 20 + 1;
		#else // OS_SYMBIAN
		int sticky = Math::Rand(seed) % 20 + 1;
		#endif	
		for (int j = 0; j < sticky; ++j)
		{
			iCards[j]->Swap(*(iCards[(52 - sticky) + j]));
		}
	}
}

int CGameCardStack::CardIndex(const CGameCard& aCard) const
{
	for (int i = 0; i < iCardCount; ++i)
	{
		if (iCards[i] == &aCard)
			return i;
	}
	return -1; // no match
}

enum ECardArrangement
{
	EIllegal = 0x00,
	ESameSuitAscending = 0x01,
	ESameSuitDescending = 0x02,
	ESameValue = 0x04,
	EAlternateColourDescending = 0x08,
	EAllowISOOS = 0x10,
	EPlayerOutHand = ESameSuitAscending|ESameSuitDescending|ESameValue,
	EAny = 0xFF,
};

// following macro helps keep matrix table narrow (source code)
#define CGCS CGameCardStack

static struct SCardMoveMatrixEntry
{
	CGCS::EType iSourceStackType;
	CGCS::EType iTargetStackType;
	int iValidStates;
	int iArrangement;
	bool iNeedSentinal; // ace required
	int iScore;
} const _CardMoveMatrix[] = 
{
	{ CGCS::EDeck,				CGCS::EDeck,			EGameStateDealing,		EAny,					false, 0 },
	{ CGCS::EDeck,				CGCS::EInStack,			EGameStateDealing,		EAny,						false, 0 },
	{ CGCS::EDeck,				CGCS::EOutStack,		EGameStateDealing,		EAny,					false, 0 },
	{ CGCS::EDeck,				CGCS::EPlayer1InHand,	EGameStateDealing,		EAny,						false, 0 },
	{ CGCS::EDeck,				CGCS::EPlayer1OutHand,	EGameStateDealing,		EAny,					false, 0 },
	{ CGCS::EDeck,				CGCS::EPlayer2InHand,	EGameStateDealing,		EAny,						false, 0 },
	{ CGCS::EDeck,				CGCS::EPlayer2OutHand,	EGameStateDealing,		EAny,					false, 0 },
	{ CGCS::EInStack,			CGCS::EDeck,			EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EInStack,			CGCS::EInStack,			EGameStateEitherPlayer,	EAlternateColourDescending, false, 0 },
	{ CGCS::EInStack,			CGCS::EOutStack,		EGameStateEitherPlayer,	ESameSuitAscending,			true, 2 },
	{ CGCS::EInStack,			CGCS::EPlayer1InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EInStack,			CGCS::EPlayer1OutHand,	EGameStatePlayer2,		EPlayerOutHand|EAllowISOOS,	true, 0 },
	{ CGCS::EInStack,			CGCS::EPlayer2InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EInStack,			CGCS::EPlayer2OutHand,	EGameStatePlayer1,		EPlayerOutHand|EAllowISOOS,	true, 0 },
	{ CGCS::EOutStack,			CGCS::EDeck,			EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EOutStack,			CGCS::EInStack,			EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EOutStack,			CGCS::EOutStack,		EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EOutStack,			CGCS::EPlayer1InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EOutStack,			CGCS::EPlayer1OutHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EOutStack,			CGCS::EPlayer2InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EOutStack,			CGCS::EPlayer2OutHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1InHand,		CGCS::EDeck,			EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1InHand,		CGCS::EOutStack,		EGameStatePlayer1,		ESameSuitAscending,			true, 2 },
	{ CGCS::EPlayer1InHand,		CGCS::EInStack,			EGameStatePlayer1,		EAlternateColourDescending, false, 0 },
	{ CGCS::EPlayer1InHand,		CGCS::EPlayer1InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1InHand,		CGCS::EPlayer1OutHand,	EGameStatePlayer1,		EAny,						false, 0 },
	{ CGCS::EPlayer1InHand,		CGCS::EPlayer2InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1InHand,		CGCS::EPlayer2OutHand,	EGameStatePlayer1,		EPlayerOutHand,				true, 1 },
	{ CGCS::EPlayer1OutHand,	CGCS::EDeck,			EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1OutHand,	CGCS::EInStack,			EGameStateEitherPlayer,	EAlternateColourDescending, false, 0 },
	{ CGCS::EPlayer1OutHand,	CGCS::EOutStack,		EGameStateEitherPlayer,	ESameSuitAscending,			true, 2 },
	{ CGCS::EPlayer1OutHand,	CGCS::EPlayer1InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1OutHand,	CGCS::EPlayer1OutHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1OutHand,	CGCS::EPlayer2InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer1OutHand,	CGCS::EPlayer2OutHand,	EGameStateEitherPlayer, EPlayerOutHand,				true, 0 },
	{ CGCS::EPlayer2InHand,		CGCS::EDeck,			EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer2InHand,		CGCS::EOutStack,		EGameStatePlayer2,		ESameSuitAscending,			true, 2 },
	{ CGCS::EPlayer2InHand,		CGCS::EInStack,			EGameStatePlayer2,		EAlternateColourDescending, false, 0 },
	{ CGCS::EPlayer2InHand,		CGCS::EPlayer2InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer2InHand,		CGCS::EPlayer2OutHand,	EGameStatePlayer2,		EAny,						false, 0 },
	{ CGCS::EPlayer2InHand,		CGCS::EPlayer1InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer2InHand,		CGCS::EPlayer1OutHand,	EGameStatePlayer2,		EPlayerOutHand,				true, 1 },
	{ CGCS::EPlayer2OutHand,	CGCS::EDeck,			EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer2OutHand,	CGCS::EInStack,			EGameStateEitherPlayer,	EAlternateColourDescending, false, 0 },
	{ CGCS::EPlayer2OutHand,	CGCS::EOutStack,		EGameStateEitherPlayer,	ESameSuitAscending,			true, 2 },
	{ CGCS::EPlayer2OutHand,	CGCS::EPlayer2InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer2OutHand,	CGCS::EPlayer2OutHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer2OutHand,	CGCS::EPlayer1InHand,	EGameStateNone,			EIllegal,					false, 0 },
	{ CGCS::EPlayer2OutHand,	CGCS::EPlayer1OutHand,	EGameStateEitherPlayer, EPlayerOutHand,				true, 0 }
};

bool CGameCardStack::CanMoveCard(EGameState aState, const CGameCard& aGameCard, const CGameCardStack& aTargetStack) const
{
	int cardIndex = CardIndex(aGameCard);
	if (cardIndex < 0)
		return false;

	if (cardIndex != iCardCount - 1)
	{
		if (Type() != EInStack || aTargetStack.Type() != EInStack)
			return false;
	}

	// first see if source and destination stack are compatible

	int theMoveMatrixIndex;
	int theMoveMatrixCount = sizeof(_CardMoveMatrix) / sizeof(_CardMoveMatrix[0]);

	for (theMoveMatrixIndex = 0; theMoveMatrixIndex < theMoveMatrixCount; ++theMoveMatrixIndex)
	{
		if (_CardMoveMatrix[theMoveMatrixIndex].iSourceStackType == iType && 
			_CardMoveMatrix[theMoveMatrixIndex].iTargetStackType == aTargetStack.iType)
		{
			if (_CardMoveMatrix[theMoveMatrixIndex].iValidStates & aState)
				break;
			else
				return false;
		}
	}

	if (theMoveMatrixIndex == theMoveMatrixCount)
		return false;

	const SCardMoveMatrixEntry& theMove = _CardMoveMatrix[theMoveMatrixIndex];

	int arrangement = theMove.iArrangement;

	if (arrangement != EAny && (arrangement & EAllowISOOS))
	{
		arrangement &= ~EAllowISOOS;
		if (!iModel.AllowISOOS())
			return false;
	}

	if (arrangement == EIllegal)
		return false;

	if (aTargetStack.CardCount() == 0)
	{
		if (theMove.iNeedSentinal)
		{
			// special rule: opponent's empty outstack space is not available for current player
			if (aTargetStack.Type() == EPlayer1OutHand || aTargetStack.Type() == EPlayer2OutHand)
				return false;
			// aces out first
			return aGameCard.Value() == CGameCard::EAce;
			// todo(maybe) - support for ordinary patience .. king sentinals in "InStacks"
		}
		return true;
	}

	CGameCard& topCard = aTargetStack.TopCard();

	if ((arrangement & ESameValue) && (aGameCard.Value() == topCard.Value()))
	{
		for (int i = cardIndex; i < iCardCount - 1; ++i)
		{
			if (iCards[i]->Value() == iCards[i+1]->Value())
				continue;
			return false;
		}
		return true;
	}

	if ((arrangement & EAlternateColourDescending) && (aGameCard.Colour() != topCard.Colour()) &&
		(topCard.Value() > CGameCard::EAce) && (aGameCard.Value() == topCard.Value() - 1))
	{
		for (int i = cardIndex; i < iCardCount - 1; ++i)
		{
			if ((iCards[i]->Colour() != iCards[i+1]->Colour()) &&
				(iCards[i+1]->Value() > CGameCard::EAce) && (iCards[i]->Value() == iCards[i+1]->Value() + 1))
				continue;
			return false;
		}
		return true;
	}

	if ((arrangement & ESameSuitAscending) && (aGameCard.Suit() == topCard.Suit()) && 
		(topCard.Value() < CGameCard::EKing) && (aGameCard.Value() == topCard.Value() + 1))
	{
		for (int i = cardIndex; i < iCardCount - 1; ++i)
		{
			if ((iCards[i]->Suit() == iCards[i+1]->Suit()) && 
				(iCards[i+1]->Value() < CGameCard::EKing) && (iCards[i]->Value() == iCards[i+1]->Value() - 1))
				continue;	
			return false;
		}
		return true;
	}

	if ((arrangement & ESameSuitDescending) && (aGameCard.Suit() == topCard.Suit()) && 
		(topCard.Value() > CGameCard::EAce) && (aGameCard.Value() == topCard.Value() - 1))
	{
		for (int i = cardIndex; i < iCardCount - 1; ++i)
		{
			if ((iCards[i]->Suit() == iCards[i+1]->Suit()) && 
				(iCards[i+1]->Value() > CGameCard::EAce) && (iCards[i]->Value() == iCards[i+1]->Value() + 1))
				continue;
			return false;
		}
		return true;
	}

	if (arrangement == EAny)
	{
		if (cardIndex == iCardCount - 1) // have to be one at a time (top card) for dealing cards or no out card move
			return true;
	}

	return false;
}

void CGameCardStack::MoveCard(EGameState aState, CGameCard& aGameCard, CGameCardStack& aTargetStack)
{
	if (!CanMoveCard(aState, aGameCard, aTargetStack))
		return;
	
	int cardIndex = CardIndex(aGameCard);

	for (int i = cardIndex; i < iCardCount; ++i)
	{
		aTargetStack.AddCard(*iCards[i]);
	}
	for (int i = iCardCount - 1; i >= cardIndex; --i)
	{
		RemoveCard(*iCards[i]);
	}
}

bool CGameCardStack::SetCard(int aIndex, CGameCard& aGameCard)
{
	if (aIndex >= iCardCount || aIndex < 0)
		return false;

	iCards[aIndex] = &aGameCard;

	return true;
}

void CGameCardStack::AddCard(CGameCard& aGameCard)
{
	if (iCardCount == 52)
		return;

	switch(iType)
	{
	case EDeck:
	case EPlayer1InHand:
	case EPlayer2InHand:
		aGameCard.SetFaceDown(true);
		break;
	default:
		aGameCard.SetFaceDown(false);
		break;
	}
		
	iCards[iCardCount++] = &aGameCard;
}

void CGameCardStack::RemoveCard(CGameCard& aGameCard)
{
	if (iCardCount == 0)
		return;

	if (iType == EDeck)
	{
		aGameCard.SetDealt(true);
		--iCardCount;
		return;
	}

	int theCardIndex = CardIndex(aGameCard);
	if (theCardIndex >= 0)
	{
		iCards[theCardIndex] = 0;

		for (int i = theCardIndex+1; i < 52; ++i)
		{
			iCards[i-1] = iCards[i];
		}
		--iCardCount;
	}
}

void CGameCardStack::SpacialInformation(int& aXPosition, int& aYPosition, int& aWidth, int& aHeight) const
{
	SSpacialInformation::SpacialInformation(aXPosition, aYPosition, aWidth, aHeight);
}

void CGameCardStack::SetSpacialInformation(int aXPosition, int aYPosition, int aWidth, int aHeight)
{
	SSpacialInformation::SetSpacialInformation(aXPosition, aYPosition, aWidth, aHeight);

	if (CardCount() == 0)
		return;

	int ignore, cardWidth, cardHeight;
	// card width and height already set, we are only interested in position in this method,
	// so keep width and height.
	static_cast<SSpacialInformation&>(*iCards[0]).SpacialInformation(ignore, ignore, cardWidth, cardHeight);

	if (iType != EInStack)
	{
		for (int i = 0; i < CardCount(); ++i)
		{
			// index be Z-order
			static_cast<SSpacialInformation&>(*iCards[i]).SetSpacialInformation(aXPosition, aYPosition, cardWidth, cardHeight);
		}
	}
	else
	{
		int maxOverlap = cardHeight / 3;
		int cardOverlap = (aHeight - ((cardHeight * 2) / 3)) / CardCount();
		if (cardOverlap > maxOverlap)
			cardOverlap = maxOverlap; 
		for (int i = 0; i < CardCount(); ++i)
		{
			static_cast<SSpacialInformation&>(*iCards[i]).SetSpacialInformation(aXPosition, aYPosition + (cardOverlap * i), cardWidth, cardHeight);
		}
	}
}
