// detail_algorithm.h
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_DETAIL_ALGORITHM
#define LIB_DETAIL_ALGORITHM

#include <cstring>
#include "type.h"

namespace lib 
{
	namespace detail
	{
		template <typename InIter, typename OutIter> inline
		OutIter copy_dispatch(InIter first, InIter last, OutIter result, false_type)
		{
			while(first != last)
				*result++ = *first++;
			return result;
		}

		template <typename T> inline
		T* copy_dispatch(const T* first, const T* last, T* result, true_type)
		{
			memmove(result, first, (last - first) * sizeof(T));
			result += (last - first);
			return result;
		}

		template <typename InIter, typename OutIter, typename T> inline
		OutIter copy(InIter first, InIter last, OutIter result, const T&)
		{
			return copy_dispatch(first, last, result, is_scalar<T>::type());
		}
		
		template <typename InIter, typename OutIter> inline
		OutIter copy_backward_dispatch(InIter first, InIter last, OutIter result, false_type)
		{
			while(first != last)
				*--result = *--last;
			return result;
		}

		template <typename T> inline
		T* copy_backward_dispatch(const T* first, const T* last, T* result, true_type)
		{
			memmove(result - (last - first), first, (last - first) * sizeof(T));
			return result - (last - first);
		}

		template <typename BidiIter1, typename BidiIter2, typename T> inline
		BidiIter2 copy_backward(BidiIter1 first, BidiIter1 last, BidiIter2 result, const T&)
		{
			return copy_backward_dispatch(first, last, result, is_scalar<T>::type());
		}

		template <typename RandIter, typename T> inline
		int partition(RandIter a, T x, int p, int r)
		{
			int i = p - 1;
			for (int j = p; j <= r - 1; j++)
			{
				if (a[j] < x)
				{
					i++;
					swap(a[i], a[j]);
				}
			}
			swap(a[i + 1], a[r]);
			return i + 1;
		}

		template <typename RandIter> inline
		void quicksort(RandIter a, int p, int r)
		{
			if (p < r)
			{
				int q = partition(a, a[r], p, r);
				quicksort(a, p, q - 1);
				quicksort(a, q + 1, r);
			}
		}

		template <typename RandIter, typename T, typename Compare> inline
		int partition(RandIter a, T x, int p, int r, Compare comp)
		{
			int i = p - 1;
			for (int j = p; j <= r - 1; j++)
			{
				if (comp(a[j], x))
				{
					i++;
					swap(a[i], a[j]);
				}
			}
			swap(a[i + 1], a[r]);
			return i + 1;
		}

		template <typename RandIter, typename Compare> inline
		void quicksort(RandIter a, int p, int r, Compare comp)
		{
			if (p < r)
			{
				int q = partition(a, a[r], p, r, comp);
				quicksort(a, p, q - 1, comp);
				quicksort(a, q + 1, r, comp);
			}
		}

		template <typename BidiIter> inline
		void bubblesort(BidiIter first, BidiIter last)
		{
			if (first == last) // empty
				return;
			--last;
			if (first == last) // just one
				return;
			bool swapped = false;
			do
			{
				BidiIter nextnext = first;
				BidiIter next = nextnext++;
				while(next != last)
				{
					if (!(*next < *nextnext))
					{
						swap(*next, *nextnext);
						swapped = true;
					}
					next++;
					nextnext++;
				}
				--last;
			} while (swapped && last != first);
		}

		template <typename BidiIter, typename Compare> inline
		void bubblesort(BidiIter first, BidiIter last, Compare comp)
		{
			if (first == last) // empty
				return;
			--last;
			if (first == last) // just one
				return;
			bool swapped = false;
			do
			{
				BidiIter nextnext = first;
				BidiIter next = nextnext++;
				while(next != last)
				{
					if (!comp(*next,*nextnext))
					{
						swap(*next, *nextnext);
						swapped = true;
					}
					next++;
					nextnext++;
				}
				--last;
			} while (swapped && last != first);
		}
	}
}

#endif // LIB_DETAIL_ALGORITHM