// detail_tree.h
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_DETAIL_TREE
#define LIB_DETAIL_TREE

#include "utility.h"

namespace lib 
{
	namespace detail
	{
		template<bool Multi, typename Key, typename T, typename KeyOfValue, typename Compare, typename A, bool Balanced = true>
		class redblack_tree
		{
		public:
			typedef Key key_type;
			typedef T value_type;
			typedef Compare key_compare;
			typedef A allocator_type;
			typedef value_type* pointer;
			typedef const value_type* const_pointer;
			typedef value_type& reference;
			typedef const value_type& const_reference;

		private:
			struct node_base;
			friend struct node_base;

			enum node_colour { BLACK, RED };

			struct node_base
			{
				node_base* iParent;
				node_base* iLeft;
				node_base* iRight;
				node_colour iColour;
				node_base() : iParent(0), iLeft(0), iRight(0), iColour(BLACK) {}
			};

			struct node : node_base
			{
				value_type iValue;
				node(const value_type& value) : iValue(value) {}
			};

			typedef node_base* node_base_pointer;
			typedef node* node_pointer;
			typedef const node_base* const_node_base_pointer;
			typedef const node* const_node_pointer;

		private:
			A iAllocator;
			key_compare Comp; 
			node_base iHeadObject;
			node_base_pointer iHead;
			node_base iNilObject;
			node_base_pointer iNil;
			size_t iSize;

		public:
			enum { node_size = sizeof(node) };

			class iterator;
			class const_iterator;
			friend class iterator;
			friend class const_iterator;

			class iterator
			{
			friend class redblack_tree<Multi, Key, T, KeyOfValue, Compare, A, Balanced>;
			friend class const_iterator;
			public:
				iterator(const iterator& x) : iHead(x.iHead), iNil(x.iNil), iNode(x.iNode) {}
				iterator() : iHead(0), iNil(0), iNode(0) {}
				iterator(node_base_pointer head, node_base_pointer nil, node_base_pointer nptr) : iHead(head), iNil(nil), iNode(nptr) {}
			private:
				node_base_pointer iHead;
				node_base_pointer iNil;
				node_base_pointer iNode;
			public:
				iterator& operator++() 
				{ 
					if (iNode == iHead)
						iNode = left(iNode);
					else
					{
						iNode = successor(iNil, iNode);
						if (iNode == iNil)
							iNode = iHead;
					}
					return *this; 
				}
				iterator& operator--() 
				{ 
					if (iNode == iHead)
						iNode = right(iNode);
					else
					{
						iNode = predecessor(iNil, iNode);
						if (iNode == iNil)
							iNode = iHead;
					}
					return *this; 
				}
				iterator operator++(int) { iterator ret(*this); operator++(); return ret; }
				iterator operator--(int) { iterator ret(*this); operator--(); return ret; }
				reference operator*() const { return static_cast<node_pointer>(iNode)->iValue; }
				pointer operator->() const { return &static_cast<node_pointer>(iNode)->iValue; }
				bool operator==(const iterator& rhs) const { return iNode == rhs.iNode; }
				bool operator!=(const iterator& rhs) const { return iNode != rhs.iNode; }
			private:
				operator node_base_pointer() const { return iNode; }
				node_base_pointer node() const { return iNode; }
			};
			class const_iterator
			{
			friend class redblack_tree<Multi, Key, T, KeyOfValue, Compare, A, Balanced>;
			friend class iterator;
			public:
				const_iterator(const const_iterator& x) : iHead(x.iHead), iNil(x.iNil), iNode(x.iNode) {}
				const_iterator(const iterator& x) : iHead(x.iHead), iNil(x.iNil), iNode(x.iNode) {}
				const_iterator() : iHead(0), iNil(0), iNode(0) {}
				const_iterator(const_node_base_pointer head, const_node_base_pointer nil, const_node_base_pointer nptr) : iHead(head), iNil(nil), iNode(nptr) {}
			private:
				const_node_base_pointer iHead;
				const_node_base_pointer iNil;
				const_node_base_pointer iNode;
			public:
				const_iterator& operator++() 
				{ 
					if (iNode == iHead)
						iNode = left(iNode);
					else
					{
						iNode = successor(iNil, iNode);
						if (iNode == iNil)
							iNode = iHead;
					}
					return *this; 
				}
				const_iterator& operator--() 
				{ 
					if (iNode == iHead)
						iNode = right(iNode);
					else
					{
						iNode = predecessor(iNil, iNode);
						if (iNode == iNil)
							iNode = iHead;
					}
					return *this; 
				}
				const_iterator operator++(int) { const_iterator ret(*this); operator++(); return ret; }
				const_iterator operator--(int) { const_iterator ret(*this); operator--(); return ret; }
				const_reference operator*() const { return static_cast<const_node_pointer>(iNode)->iValue; }
				const_pointer operator->() const { return &static_cast<const_node_pointer>(iNode)->iValue; }
				bool operator==(const const_iterator& rhs) const { return iNode == rhs.iNode; }
				bool operator!=(const const_iterator& rhs) const { return iNode != rhs.iNode; }
			private:
				operator const_node_base_pointer() const { return iNode; }
				const_node_base_pointer node() const { return iNode; }
			};

		public:
			// construction
			redblack_tree(const A& allocator = A()) : iAllocator(allocator), iHead(&iHeadObject), iNil(&iNilObject), iSize(0) 
			{
				left(iHead) = iHead;
				right(iHead) = iHead;
				parent(iHead) = iNil;
				left(iNil) = iNil;
				right(iNil) = iNil;
				parent(iNil) = iNil;
			}
			~redblack_tree() 
			{
				while(!empty())
					erase(begin());
			}

			// traversals
			iterator begin() { return iterator(iHead, iNil, left(iHead)); }
			const_iterator begin() const { return const_iterator(iHead, iNil, left(iHead)); }
			iterator end() { return iterator(iHead, iNil, iHead); }
			const_iterator end() const { return const_iterator(iHead, iNil, iHead); }
			bool empty() const { return iHead->iLeft == iHead; }
			size_t depth() const { return depth(root()); }
			size_t depth(const_node_base_pointer ptr) const
			{
				if (ptr != iNil)
				{
					size_t depthLeft = depth(ptr->iLeft);
					size_t depthRight = depth(ptr->iRight);
					if (depthLeft > depthRight)
						return 1 + depthLeft;
					else
						return 1 + depthRight;
				}
				return 0;
			}
			// element access
			iterator lower_bound(const key_type& k)
			{
				node_base_pointer next = root();
				node_base_pointer not_less_than = iHead;
				while(next != iNil)
				{
					if (!Comp(key(next), k)) // next >= k
					{
						not_less_than = next;
						next = left(next);
					}
					else // next < k
						next = right(next);
				}
				return iterator(iHead, iNil, not_less_than);
			}
			const_iterator lower_bound(const key_type& k) const
			{
				const_node_base_pointer next = root();
				const_node_base_pointer not_less_than = iHead;
				while(next != iNil)
				{
					if (!Comp(key(next), k)) // next >= k
					{
						not_less_than = next;
						next = left(next);
					}
					else // next < k
						next = right(next);
				}
				return const_iterator(iHead, iNil, not_less_than);
			}
			iterator upper_bound(const key_type& k)
			{
				node_base_pointer next = root();
				node_base_pointer greater_than = iHead;
				while(next != iNil)
				{
					if (Comp(k, key(next))) // next > k
					{
						greater_than = next;
						next = left(next);
					}
					else // next <= k
						next = right(next);
				}
				return iterator(iHead, iNil, greater_than);
			}
			const_iterator upper_bound(const key_type& k) const
			{
				const_node_base_pointer next = root();
				const_node_base_pointer greater_than = iHead;
				while(next != iNil)
				{
					if (Comp(k, key(next))) // next > k
					{
						greater_than = next;
						next = left(next);
					}
					else // next <= k
						next = right(next);
				}
				return const_iterator(iHead, iNil, greater_than);
			}
			// modifiers
			pair<iterator, bool> insert(const value_type& value) 
			{			 
				node_base_pointer new_node = 0;
				node_base_pointer y = iNil;
				node_base_pointer x = root();
				while(x != iNil)
				{
					y = x;
					if (Comp(KeyOfValue()(value), key(x)))
						x = left(x);
					else if (Comp(key(x), KeyOfValue()(value)))
						x = right(x);
					else
						return make_pair(iterator(iHead, iNil, x), false);
				}
				new_node = buy_node(value);
				parent(new_node) = y;
				if (y == iNil)
					root() = new_node;
				else if (Comp(key(new_node), key(y)))
					left(y) = new_node;
				else
					right(y) = new_node;
				colour(new_node) = RED;
				insert_fixup(new_node);
				iSize++;
				if (left(iHead) == iHead || Comp(key(new_node), key(left(iHead))))
					left(iHead) = new_node;
				if (right(iHead) == iHead || Comp(key(right(iHead)), key(new_node)))
					right(iHead) = new_node;
				return make_pair(iterator(iHead, iNil, new_node), true);
			}
			void insert_fixup(node_base_pointer z)
			{
				if (!Balanced)
					return;

				while(colour(parent(z)) == RED)
				{
					if (parent(z) == left(parent(parent(z))))
					{
						node_base_pointer y = right(parent(parent(z)));
						if (colour(y) == RED)
						{
							colour(parent(z)) = BLACK;
							colour(y) = BLACK;
							colour(parent(parent(z))) = RED;
							z = parent(parent(z));
						}
						else
						{
							if (z == right(parent(z)))
							{
								z = parent(z);
								left_rotate(z);
							}
							colour(parent(z)) = BLACK;
							colour(parent(parent(z))) = RED;
							right_rotate(parent(parent(z)));
						}
					}
					else
					{
						node_base_pointer y = left(parent(parent(z)));
						if (colour(y) == RED)
						{
							colour(parent(z)) = BLACK;
							colour(y) = BLACK;
							colour(parent(parent(z))) = RED;
							z = parent(parent(z));
						}
						else
						{
							if (z == left(parent(z)))
							{
								z = parent(z);
								right_rotate(z);
							}
							colour(parent(z)) = BLACK;
							colour(parent(parent(z))) = RED;
							left_rotate(parent(parent(z)));
						}
					}
				}
				colour(root()) = BLACK;
			}
			void erase(iterator position) 
			{ 
				node_base_pointer z = position.node();
				node_base_pointer y;
				if (left(z) == iNil || right(z) == iNil)
					y = z;
				else
					y = successor(iNil, z);
				node_base_pointer x;
				if (left(y) != iNil)
					x = left(y);
				else
					x = right(y);
				parent(x) = parent(y);
				if (parent(y) == iNil)
				{
					root() = x;
					parent(root()) = iNil;
				}
				else
					if (y == left(parent(y)))
						left(parent(y)) = x;
					else
						right(parent(y)) = x;
				node_colour y_colour = colour(y);
				if (y != z)
				{
					parent(y) = parent(z);
					left(y) = left(z);
					right(y) = right(z);
					colour(y) = colour(z);
					if (parent(y) != iNil && left(parent(y)) == z)
						left(parent(y)) = y;
					else if (parent(y) != iNil)
						right(parent(y)) = y;
					if (left(y) != iNil)
						parent(left(y)) = y;
					if (right(y) != iNil)
						parent(right(y)) = y;
					if (parent(y) == iNil && y != root())
						root() = y;
				}
				--iSize;
				assert(x != z);
				if (y_colour == BLACK)
					delete_fixup(x);

				if (root() != iNil)
				{
					if (iHead->iLeft == z)
						iHead->iLeft = leftmost(iNil, root());
					if (iHead->iRight == z)
						iHead->iRight = rightmost(iNil, root());
				}
				else
				{
					iHead->iLeft = iHead;
					iHead->iRight = iHead;
				}
				assert(z != iNil);
				iAllocator.destruct(z);
				iAllocator.deallocate(reinterpret_cast<A::pointer>(z));
			}
			void delete_fixup(node_base_pointer x)
			{
				if (!Balanced)
					return;
				
				while (x != iNil && x != root() && colour(x) == BLACK)
				{
					if (x == left(parent(x)))
					{
						node_base_pointer w = right(parent(x));
						if (colour(w) == RED)
						{
							colour(w) = BLACK;
							colour(parent(x)) = RED;
							left_rotate(parent(x));
							w = right(parent(x));
						}
						if (colour(left(w)) == BLACK && colour(right(w)) == BLACK)
						{
							colour(w) = RED;
							x = parent(x);
						}
						else
						{
							if (colour(right(w)) == BLACK)
							{
								colour(left(w)) = BLACK;
								colour(w) = RED;
								right_rotate(w);
								w = right(parent(x));
							}
							colour(w) = colour(parent(x));
							colour(parent(x)) = BLACK;
							colour(right(w)) = BLACK;
							left_rotate(parent(x));
							x = root();
						}
					}
					else
					{
						node_base_pointer w = left(parent(x));
						if (colour(w) == RED)
						{
							colour(w) = BLACK;
							colour(parent(x)) = RED;
							right_rotate(parent(x));
							w = left(parent(x));
						}
						if (colour(right(w)) == BLACK && colour(left(w)) == BLACK)
						{
							colour(w) = RED;
							x = parent(x);
						}
						else
						{
							if (colour(left(w)) == BLACK)
							{
								colour(right(w)) = BLACK;
								colour(w) = RED;
								left_rotate(w);
								w = left(parent(x));
							}
							colour(w) = colour(parent(x));
							colour(parent(x)) = BLACK;
							colour(left(w)) = BLACK;
							right_rotate(parent(x));
							x = root();
						}
					}
					left(iNil) = iNil;
					right(iNil) = iNil;
					parent(iNil) = iNil;
					colour(iNil) = BLACK;
				}
				colour(x) = BLACK;
			}
		private:
			// implementation
			static const Key& key(node_base_pointer current) { return KeyOfValue()(static_cast<node_pointer>(current)->iValue); }
			node_base_pointer buy_node(const value_type& value)
			{
				node_pointer new_node  = reinterpret_cast<node_pointer>(iAllocator.allocate(1, sizeof(node)));
				iAllocator.construct(new_node, node(value));
				parent(new_node) = iNil;
				left(new_node) = iNil;
				right(new_node) = iNil;
				return new_node;
			}
			node_base_pointer& root() { return iHead->iParent; }
			node_base_pointer leftmost() { return iHead->iLeft; }
			static node_base_pointer leftmost(node_base_pointer nil, node_base_pointer subtree) 
			{
				node_base_pointer next = subtree;
				while(left(next) != nil)
					next = left(next);
				return next;
			}
			node_base_pointer rightmost() { return iHead->iRight; } 
			static node_base_pointer rightmost(node_base_pointer nil, node_base_pointer subtree) 
			{ 
				node_base_pointer next = subtree;
				while(right(next) != nil)
					next = right(next);
				return next;
			}
			static node_base_pointer successor(node_base_pointer nil, node_base_pointer x)
			{
				if (right(x) != nil)
					return leftmost(nil, right(x));
				node_base_pointer y = parent(x);
				while(y != nil && x == right(y))
				{
					x = y;
					y = parent(y);
				}
				return y;
			}
			static node_base_pointer predecessor(node_base_pointer nil, node_base_pointer x)
			{
				if (left(x) != nil)
					return rightmost(nil, left(x));
				node_base_pointer y = parent(x);
				while(y != nil && x == left(y))
				{
					x = y;
					y = parent(y);
				}
				return y;
			}
			void left_rotate(node_base_pointer x)
			{
				node_base_pointer y = right(x);
				right(x) = left(y);
				if (left(y) != iNil)
					parent(left(y)) = x;
				assert(y == iNil || x != iNil);
				parent(y) = parent(x);
				if (parent(x) == iNil)
					root() = y;
				else
					if (x == left(parent(x)))
						left(parent(x)) = y;
					else
						right(parent(x)) = y;
				left(y) = x;
				parent(x) = y;
			}
			void right_rotate(node_base_pointer x)
			{
				node_base_pointer y = left(x);
				left(x) = right(y);
				if (right(y) != iNil)
					parent(right(y)) = x;
				assert(y == iNil || x != iNil);
				parent(y) = parent(x);
				if (parent(x) == iNil)
					root() = y;
				else
					if (x == right(parent(x)))
						right(parent(x)) = y;
					else
						left(parent(x)) = y;
				right(y) = x;
				parent(x) = y;
			}
			static node_base_pointer& left(node_base_pointer current) { return current->iLeft; }
			static node_base_pointer& right(node_base_pointer current) { return current->iRight; }
			static node_base_pointer& parent(node_base_pointer current) { return current->iParent; }
			static node_colour& colour(node_base_pointer current) { return current->iColour; }
			static const Key& key(const_node_base_pointer current) { return KeyOfValue()(static_cast<const_node_pointer>(current)->iValue); }
			const_node_base_pointer root() const { return iHead->iParent; }
			const_node_base_pointer leftmost() const { return iHead->iLeft; }
			static const_node_base_pointer leftmost(const_node_base_pointer nil, const_node_base_pointer subtree) 
			{
				const_node_base_pointer next = subtree;
				while(left(next) != nil)
					next = left(next);
				return next;
			}
			const_node_base_pointer rightmost() const { return iHead->iRight; }
			static const_node_base_pointer rightmost(const_node_base_pointer nil, const_node_base_pointer subtree)
			{ 
				const_node_base_pointer next = subtree;
				while(right(next) != nil)
					next = right(next);
				return next;
			}
			static const_node_base_pointer successor(const_node_base_pointer nil, const_node_base_pointer x)
			{
				if (right(x) != nil)
					return leftmost(nil, right(x));
				const_node_base_pointer y = parent(x);
				while(y != nil && x == right(y))
				{
					x = y;
					y = parent(y);
				}
				return y;
			}
			static const_node_base_pointer predecessor(const_node_base_pointer nil, const_node_base_pointer x)
			{
				if (left(x) != nil)
					return rightmost(nil, left(x));
				const_node_pointer y = parent(x);
				while(y != nil && x == left(y))
				{
					x = y;
					y = parent(y);
				}
				return y;
			}
			static const_node_base_pointer left(const_node_base_pointer current) { return current->iLeft; }
			static const_node_base_pointer right(const_node_base_pointer current) { return current->iRight; }
			static const_node_base_pointer parent(const_node_base_pointer current) { return current->iParent; }
			static node_colour colour(const_node_base_pointer current) { return current->iColour; }

			void INVARIANT_CHECK()
			{
				size_t total = INVARIANT_CHECK(root());
				assert(total == iSize);
			}
			size_t INVARIANT_CHECK(node_base_pointer nptr)
			{
				if (nptr != iNil)
				{
					size_t ret = 1;
					if (left(nptr) != iNil)
						ret += INVARIANT_CHECK(left(nptr));
					assert(left(nptr) == iNil || left(nptr) != right(nptr));
					assert(parent(nptr) == iNil || left(nptr) == iNil || parent(nptr) != left(nptr));
					assert(parent(nptr) == iNil || right(nptr) == iNil || parent(nptr) != right(nptr));
					if (right(nptr) != iNil)
						ret += INVARIANT_CHECK(right(nptr));
					return ret;
				}
				else
					return 0;
			}
		};
	}
}

#endif // LIB_DETAIL_TREE