// list.h
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

/* WARNING: This class is not a substitute for the std::list available on your platform
 * which you should be using instead.  It is here either for technical reasons or for when 
 * there is no standard library available.
  */

#ifndef LIB_LIST
#define LIB_LIST

#include "memory.h"
#include "type.h"

namespace lib 
{
	template<typename T, typename A = allocator<T> >
	class list
	{
	private:
		struct node_base
		{
			node_base* iNext;
			node_base* iPrev;
			node_base() : iNext(0), iPrev(0) {}
			void swap(node_base& x)
			{
				lib::swap(iNext, x.iNext);
				lib::swap(iPrev, x.iPrev);
				if (iNext == &x)
					iNext = this;
				if (iPrev == &x)
					iPrev = this;
				if (x.iNext == this)
					x.iNext = &x;
				if (x.iPrev == this)
					x.iPrev = &x;
				iNext->iPrev = this;
				iPrev->iNext = this;
				x.iNext->iPrev = &x;
				x.iPrev->iNext = &x;
			}
		};

		struct node : node_base
		{
			T iValue;
		};

		typedef node_base* node_base_pointer;
		typedef node* node_pointer;
		typedef const node_base* const_node_base_pointer;
		typedef const node* const_node_pointer;

	public:
		typedef T value_type;
		typedef T* pointer;
		typedef const T* const_pointer;
		typedef T& reference;
		typedef const T& const_reference;
		typedef A allocator_type;
		typedef size_t size_type;
		typedef ptrdiff_t difference_type;
		enum { node_size = sizeof(node) };

		class iterator;
		class const_iterator;

		class iterator : public std::iterator<std::bidirectional_iterator_tag, value_type, difference_type>
		{
		friend class list<T, A>;
		friend class const_iterator;
		public:
			iterator() : iNode(0) {}
			iterator(const iterator& x) : iNode(x.iNode) {}
		private:
			iterator(node_base_pointer nptr) : iNode(nptr) {}
		private:
			node_base_pointer iNode;
		public:
			iterator& operator++() { iNode = iNode->iNext; return *this; }
			iterator& operator--() { iNode = iNode->iPrev; return *this; }
			iterator operator++(int) { iterator ret(*this); operator++(); return ret; }
			iterator operator--(int) { iterator ret(*this); operator--(); return ret; }
			reference operator*() const { return static_cast<node_pointer>(iNode)->iValue; }
			pointer operator->() const { return &static_cast<node_pointer>(iNode)->iValue; }
			bool operator==(const iterator& rhs) const { return iNode == rhs.iNode; }
			bool operator!=(const iterator& rhs) const { return iNode != rhs.iNode; }
		};
		class const_iterator : public std::iterator<std::bidirectional_iterator_tag, value_type, difference_type>
		{
		friend class list<T, A>;
		friend class iterator;
		public:
			const_iterator() : iNode(0) {}
			const_iterator(const const_iterator& x) : iNode(x.iNode) {}
			const_iterator(const iterator& x) : iNode(x.iNode) {}
		private:
			const_iterator(const_node_base_pointer aNode) : iNode(aNode) {}
		private:
			const_node_base_pointer iNode;
		public:
			const_iterator& operator++() { iNode = iNode->iNext; return *this; }
			const_iterator& operator--() { iNode = iNode->iPrev; return *this; }
			const_iterator operator++(int) { const_iterator ret(*this); operator++(); return ret; }
			const_iterator operator--(int) { const_iterator ret(*this); operator--(); return ret; }
			const_reference operator*() const { return static_cast<const_node_pointer>(iNode)->iValue; }
			const_pointer operator->() const { return &static_cast<const_node_pointer>(iNode)->iValue; }
			bool operator==(const const_iterator& rhs) const { return iNode == rhs.iNode; }
			bool operator!=(const const_iterator& rhs) const { return iNode != rhs.iNode; }
		};

	public:
		// construction
		list() : iSize(0)
		{
			iRoot.iNext = &iRoot;
			iRoot.iPrev = &iRoot;
		}
		list(const A& allocator) : iAllocator(allocator), iSize(0)
		{
			iRoot.iNext = &iRoot;
			iRoot.iPrev = &iRoot;
		}
		list(const list& rhs) : iAllocator(rhs.iAllocator), iSize(0)
		{
			iRoot.iNext = &iRoot;
			iRoot.iPrev = &iRoot;
			*this = rhs;
		}
		~list()
		{
			while(!empty())
				pop_back();
		}
		// traversals
		iterator begin() { return iterator(iRoot.iNext); }
		const_iterator begin() const { return const_iterator(iRoot.iNext); }
		iterator end() { return iterator(&iRoot); }
		const_iterator end() const { return const_iterator(&iRoot); }
		bool empty() const { return iRoot.iNext == &iRoot; }
		size_type size() const { return iSize; }
		// element access
		reference front() { return *begin(); }
		const_reference front() const { return *begin(); }
		reference back() { return *--end(); }
		const_reference back() const { return *--end(); }
		// modifiers
		iterator insert(iterator position, const T& value)
		{
			node_base_pointer prev = position.iNode->iPrev;
			node_pointer new_node = reinterpret_cast<node_pointer>(iAllocator.allocate(1, sizeof(node)));
			iAllocator.construct(new_node, node());
			new_node->iValue = value;
			new_node->iPrev = prev;
			new_node->iNext = prev->iNext;
			new_node->iNext->iPrev = new_node;
			prev->iNext = new_node;

			++iSize;

			return iterator(new_node);
		}
		iterator erase(iterator position) 
		{ 
			if (position == end()) return position; 
			position.iNode->iPrev->iNext = position.iNode->iNext;
			position.iNode->iNext->iPrev = position.iNode->iPrev;
			iterator ret(position.iNode->iNext);
			iAllocator.destruct(static_cast<node_pointer>(position.iNode));
			iAllocator.deallocate(reinterpret_cast<A::pointer>(position.iNode));
			--iSize;
			return ret;
		}
		void erase(iterator first, iterator last) { while(first != last) erase(first++); }
		void push_front(const T& value)
		{
			insert(begin(), value);
		}
		void pop_front()
		{
			erase(begin());
		}
		void push_back(const T& value)
		{
			insert(end(), value);
		}
		void pop_back()
		{
			erase(--end());
		}
		void remove(const T& value, bool multiple = true)
		{
			for (iterator i = begin(); i != end(); )
			{
				if (*i == value)
				{
					erase(i++);
					if (!multiple)
						return;
				}
				else
					i++;
			}
		}
		list& operator=(const list& aOther)
		{
			clear();
			for(const_iterator i = aOther.begin(); i != aOther.end(); ++i)
				push_back(*i);
			return *this;
		}
		void clear() { erase(begin(), end()); }
		void swap(list& x)
		{
			if (iAllocator != x.iAllocator)
				lib::swap(iAllocator, x.iAllocator);
			iRoot.swap(x.iRoot);
			lib::swap(iSize, x.iSize);
		}

	private:
		// attributes
		A iAllocator;
		node_base iRoot;
		size_type iSize;
	};
}

#endif // LIB_LIST