// map.h
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

/* WARNING: This class is not a substitute for the std::map available on your platform
 * which you should be using instead.  It is here either for technical reasons or for when 
 * there is no standard library available.
  */

#ifndef LIB_MAP
#define LIB_MAP

#include "memory.h"
#include "utility.h"
#include "functional.h"
#include "detail_tree.h"

namespace lib 
{
	template<typename Key, typename T, typename Compare = less<Key>, typename A = allocator<pair<const Key, T> > >
	class map
	{
	public:
		typedef Key key_type;
		typedef T mapped_type;
		typedef pair<const Key, T> value_type;
	private:
		struct KeyFunc
		{
			const Key& operator()(const value_type& value) const { return value.first; }
		};
		typedef detail::redblack_tree<false, Key, value_type, KeyFunc, Compare, A> tree_impl;
	public:
		typedef typename tree_impl::pointer pointer;
		typedef typename tree_impl::const_pointer const_pointer;
		typedef typename tree_impl::reference reference;
		typedef typename tree_impl::const_reference const_reference;
		typedef typename tree_impl::allocator_type allocator_type;
		typedef typename tree_impl::iterator iterator;
		typedef typename tree_impl::const_iterator const_iterator;

		enum { node_size = tree_impl::node_size };

	private:
		tree_impl iTree;

	public:
		// construction
		map(const A& allocator = A()) : iTree(allocator) {}
		~map() {}

		// traversals
		iterator begin() { return iTree.begin(); }
		const_iterator begin() const { return iTree.begin(); }
		iterator end() { return iTree.end(); }
		const_iterator end() const { return iTree.end(); }
		iterator lower_bound(const key_type& x) { return iTree.lower_bound(x); }
		const_iterator lower_bound(const key_type& x) const { return iTree.lower_bound(x); }
		iterator upper_bound(const key_type& x) { return iTree.upper_bound(x); }
		const_iterator upper_bound(const key_type& x) const { return iTree.upper_bound(x); }
		bool empty() const { return iTree.empty(); }
		size_t tree_depth() const { return iTree.depth(); }
		// element access
		T& operator[](const key_type& x) { return (*((insert(make_pair(x, T()))).first)).second; }
		// modifiers
		pair<iterator, bool> insert(const value_type& value) { return iTree.insert(value); }
		void erase(iterator position) { iTree.erase(position); }
		size_t erase(const key_type& x) 
		{ 
			iterator first = iTree.lower_bound(x);
			iterator last = iTree.upper_bound(x);
			size_t count = 0;
			while(first != last)
				erase(first++), count++;
			return count;
		}
		void erase(iterator first, iterator last) { while(first != last) erase(first++); }
		void remove(const T& value)
		{
			for (iterator i = begin(); i != end(); )
			{
				if (*i == value)
					erase(i++);
				else
					i++;
			}
		}
	};
}

#endif // LIB_MAP