// memory.h
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

/* WARNING: The classes present here are not a substitute for any equivalent std:: 
 * classes available on your platform which you should be using instead.  They exist here
 * either for technical reasons or for when there is no standard library available.
  */

#ifndef LIB_MEMORY
#define LIB_MEMORY

#include "detail_memory.h"

namespace lib
{
	template <typename InIter, typename OutIter> inline
	OutIter uninitialized_copy(InIter first, InIter last, OutIter result)
	{
		return detail::uninitialized_copy(first, last, result, *result);
	}

	template <typename T>
	class allocator
	{
	public:
		typedef T value_type;
		typedef T* pointer;
		typedef T& reference;
		typedef const T* const_pointer;
		typedef const T& const_reference;
		typedef size_t size_type;
		typedef ptrdiff_t difference_type;

	// construction
	public:
		allocator() {}
		allocator(const allocator& aAllocator) {}
		allocator& operator=(const allocator&) { return *this; }

	// operations
	public:
		// need aObjectSize as VC++6 does not support rebind
		T* allocate(size_type aCount = 1, size_t aObjectSize = sizeof(T))
		{
			return static_cast<T*>(::operator new (aCount * aObjectSize));
		}

		void deallocate(T* aObject, size_type = 1)
		{
			::operator delete(static_cast<void*>(aObject));
		}
		
		template <typename T2>
		void construct(T2* aObject, const T2& val)
		{
			new (static_cast<void*>(aObject)) T2(val);
		}
		
		template <typename T2>
		void destruct(T2* aObject)
		{
			aObject->~T2();
		}

		template <typename T2>
		void destroy(T2* aObject)
		{
			destruct(aObject);
			deallocate(reinterpret_cast<pointer>(aObject));
		}

		bool operator==(const allocator&) { return true; }
		bool operator!=(const allocator&) { return false; }

		template <typename U>
		struct rebind
		{
			typedef allocator<U> other;
		};
	};

	template <typename T, size_t N = sizeof(T), size_t ChunkSize = 8 * 1024 - 16>
	class chunk_allocator
	{
	public:
		typedef T value_type;
		typedef T* pointer;
		typedef T& reference;
		typedef const T* const_pointer;
		typedef const T& const_reference;
		typedef size_t size_type;
		typedef ptrdiff_t difference_type;

	// implementation
	private:
		struct link { link* iNext; };
		struct chunk
		{
			enum { size = ChunkSize};
			char iMem[size];
			chunk* iNext;
		};

	// construction
	public:
		chunk_allocator() : iElementSize(N < sizeof(link) ? sizeof(link) : N), iOwnChunks(0), iOwnHead(0), iChunks(&iOwnChunks), iHead(&iOwnHead) {}
		chunk_allocator(const chunk_allocator& rhs) : iElementSize(rhs.iElementSize), iOwnChunks(0), iOwnHead(0), iChunks(const_cast<chunk_allocator&>(rhs).iChunks), iHead(const_cast<chunk_allocator&>(rhs).iHead) {}
		~chunk_allocator()
		{
			if (&iOwnChunks == iChunks)
			{
				chunk* n = *iChunks;
				while(n)
				{
					chunk* p = n;
					n = n->iNext;
					delete p;
				}
			}
			else
			{
			}
		}

	// operations
	public:
		T* allocate(size_type aCount = 1, size_t = N)
		{
			assert(aCount == 1);
			if (*iHead == 0)
				grow();
			link* p = *iHead;
			*iHead = p->iNext;
			return reinterpret_cast<T*>(p);
		}

		void deallocate(T* aObject, size_type aCount = 1)
		{
			assert(aCount == 1);
			link* p = reinterpret_cast<link*>(aObject);
			p->iNext = *iHead;
			*iHead = p;
		}

		template <typename T2>
		void construct(T2* aObject, const T2& val)
		{
			new (static_cast<void*>(aObject)) T2(val);
		}
		
		template <typename T2>
		void destruct(T2* aObject)
		{
			aObject->~T2();
		}

		template <typename T2>
		void destroy(T2* aObject)
		{
			destruct(aObject);
			deallocate(reinterpret_cast<pointer>(aObject));
		}

		bool operator==(const chunk_allocator& x) const { return iChunks == x.iChunks && iHead == x.iHead; }
		bool operator!=(const chunk_allocator& x) const { return iChunks != x.iChunks || iHead != x.iHead; }

		template <typename U>
		struct rebind
		{
			typedef chunk_allocator<U, sizeof(U), ChunkSize> other;
		};

	private:
		void grow()
		{
			chunk* n = new chunk;
			n->iNext = *iChunks;
			*iChunks = n;

			const int nelem = chunk::size / iElementSize;
			char* start = n->iMem;
			char* last = &start[(nelem-1) * iElementSize];
			for (char* p = start; p < last; p+= iElementSize)
				reinterpret_cast<link*>(p)->iNext = reinterpret_cast<link*>(p+iElementSize);
			reinterpret_cast<link*>(last)->iNext = 0;
			*iHead = reinterpret_cast<link*>(start);
		}

	// attributes
	private:
		size_type iElementSize;
		chunk* iOwnChunks;
		link* iOwnHead;
		chunk** iChunks;
		link** iHead;
	};

	template <typename T>
	class platform_allocator
	{
	public:
		typedef T value_type;
		typedef T* pointer;
		typedef T& reference;
		typedef const T* const_pointer;
		typedef const T& const_reference;
		typedef size_t size_type;
		typedef ptrdiff_t difference_type;

	// construction
	public:
		platform_allocator() {}
		platform_allocator(const platform_allocator& aAllocator) {}
		platform_allocator& operator=(const platform_allocator&) { return *this; }

	// operations
	public:
		// need aObjectSize as VC++6 does not support rebind
		T* allocate(size_type aCount = 1, size_t aObjectSize = sizeof(T))
		{
			return static_cast<T*>(::VirtualAlloc(NULL, aCount * aObjectSize, MEM_RESET, PAGE_READWRITE));
		}

		void deallocate(T* aObject, size_type = 1)
		{
			::VirtualFree(aObject, 0, MEM_RELEASE);
		}
		
		template <typename T2>
		void construct(T2* aObject, const T2& val)
		{
			new (static_cast<void*>(aObject)) T2(val);
		}
		
		template <typename T2>
		void destruct(T2* aObject)
		{
			aObject->~T2();
		}

		template <typename T2>
		void destroy(T2* aObject)
		{
			destruct(aObject);
			deallocate(reinterpret_cast<pointer>(aObject));
		}

		bool operator==(const platform_allocator&) { return true; }
		bool operator!=(const platform_allocator&) { return false; }

		template <typename U>
		struct rebind
		{
			typedef platform_allocator<U> other;
		};
	};
}

#endif // LIB_MEMORY