// network.cpp - v1.0
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#include "stdafx.h"
#include "network.h"
#include "serialization.h"
#include "utility.h"

namespace lib
{
	packet::packet() : iId(0), iLength(0), iAllocLength(0), iData(0)
	{
	}

	packet::packet(const packet& aPacket) : iId(aPacket.iId), iLength(aPacket.iLength), iAllocLength(aPacket.iLength), iData(iLength != 0 ? new char[iLength] : 0)
	{
		if (iData)
			memcpy(iData, aPacket.iData, iLength);
	}

	packet::packet(int aId) : iId(aId), iLength(0), iAllocLength(0), iData(0)
	{
	}

	packet::packet(int aId, size_t aLength, const void* aData) : iId(aId), iLength(aLength), iAllocLength(aLength), iData(iLength != 0 ? new char[iLength] : 0)
	{
		if (iData)
			memcpy(iData, aData, iLength);
	}

	packet::packet(int aId, char aData) : iId(aId), iLength(sizeof(char)), iAllocLength(sizeof(char)), iData(new char[sizeof(char)])
	{
		memcpy(iData, &aData, sizeof(char));
	}

	packet::packet(int aId, int aData) : iId(aId), iLength(sizeof(int)), iAllocLength(sizeof(int)), iData(new char[sizeof(int)])
	{
		memcpy(iData, &aData, sizeof(int));
	}

	packet::packet(size_t aLength, const void* aData) : iId(0), iLength(aLength), iAllocLength(aLength), iData(iLength != 0 ? new char[iLength] : 0)
	{
		if (iData)
			memcpy(iData, aData, iLength);
	}

	packet::~packet()
	{
		if (iData != 0)
			delete[] static_cast<char*>(iData);
	}

	void packet::set(int aId, char aData)
	{
		set(aId, sizeof(char), &aData);
	}

	void packet::set(int aId, int aData)
	{
		set(aId, sizeof(int), &aData);
	}

	void packet::set(int aId, size_t aLength, const void* aData)
	{
		if (iData != 0 && iAllocLength < aLength)
		{
			delete[] static_cast<char*>(iData);
			iData = 0;
		}

		iId = aId;
		iLength = aLength;
		if (iLength != 0)
		{
			if (iData == 0)
			{
				iData = new char[iLength];
				iAllocLength = iLength;
			}
			if (aData)
				memcpy(iData, aData, iLength);
		}
	}

	void packet::set(size_t aLength, const void* aData)
	{
		set(0, aLength, aData);
	}

	template <>
	std::string packet_value<std::string>(const packet& aPacket)
	{
		std::string ret;
		ret.assign(reinterpret_cast<const char*>(aPacket.data()), aPacket.length());
		return ret;
	}

	void save_to_packet(packet& aPacket, serializable& aObject)
	{
		size_t length;
		aObject.save(lib::len_outstream(length));
		aPacket.set(aPacket.id(), length, 0);
		aObject.save(lib::ptr_outstream(aPacket.data(), length));
	}

	void load_from_packet(const packet& aPacket, serializable& aObject)
	{
		aObject.load(lib::ptr_instream(aPacket.data(), aPacket.length()));
	}

	struct socket_error
	{
		int iErrorCode;
		const char* iErrorString;
	} sSocketErrors[] =
	{
		{ WSAEACCES, "Permission denied." },
		{ WSAEADDRINUSE, "Address already in use." },
		{ WSAEADDRNOTAVAIL, "Cannot assign requested address." },
		{ WSAEAFNOSUPPORT, "Address family not supported by protocol family." },
		{ WSAEALREADY, "Operation already in progress." },
		{ WSAECONNABORTED, "Software caused connection abort." },
		{ WSAECONNREFUSED, "Connection refused." },
		{ WSAECONNRESET, "Connection reset by peer." },
		{ WSAEDESTADDRREQ, "Destination address required." },
		{ WSAEFAULT, "Bad address." },
		{ WSAEHOSTDOWN, "Host is down." },
		{ WSAEHOSTUNREACH, "No route to host." },
		{ WSAEINPROGRESS, "Operation now in progress." },
		{ WSAEINTR, "Interrupted function call." },
		{ WSAEINVAL, "Invalid argument." },
		{ WSAEISCONN, "Socket is already connected." },
		{ WSAEMFILE, "Too many open files." },
		{ WSAEMSGSIZE, "Message too long." },
		{ WSAENETDOWN, "Network is down." },
		{ WSAENETRESET, "Network dropped connection on reset." },
		{ WSAENETUNREACH, "Network is unreachable." },
		{ WSAENOBUFS, "No buffer space available." },
		{ WSAENOPROTOOPT, "Bad protocol option." },
		{ WSAENOTCONN, "Socket is not connected." },
		{ WSAENOTSOCK, "Socket operation on non-socket." },
		{ WSAEOPNOTSUPP, "Operation not supported." },
		{ WSAEPFNOSUPPORT, "Protocol family not supported." },
		{ WSAEPROCLIM, "Too many processes." },
		{ WSAEPROTONOSUPPORT, "Protocol not supported." },
		{ WSAEPROTOTYPE, "Protocol wrong type for socket." },
		{ WSAESHUTDOWN, "Cannot send after socket shutdown." },
		{ WSAESOCKTNOSUPPORT, "Socket type not supported." },
		{ WSAETIMEDOUT, "Connection timed out." },
		{ WSATYPE_NOT_FOUND, "Class type not found." },
		{ WSAEWOULDBLOCK, "Resource temporarily unavailable." },
		{ WSAHOST_NOT_FOUND, "Host not found." },
		{ WSA_INVALID_HANDLE, "Specified event object handle is invalid." },
		{ WSA_INVALID_PARAMETER, "One or more parameters are invalid." },
		{ WSAEINVALIDPROCTABLE, "Invalid procedure table from service provider." },
		{ WSAEINVALIDPROVIDER, "Invalid service provider version number." },
		{ WSA_IO_INCOMPLETE, "Overlapped I/O event object not in signaled state." },
		{ WSA_IO_PENDING, "Overlapped operations will complete later." },
		{ WSA_NOT_ENOUGH_MEMORY, "Insufficient memory available." },
		{ WSANOTINITIALISED, "Successful WSAStartup not yet performed." },
		{ WSANO_DATA, "Valid name, no data record of requested type." },
		{ WSANO_RECOVERY, "This is a non-recoverable error." },
		{ WSAEPROVIDERFAILEDINIT, "Unable to initialize a service provider." },
		{ WSASYSCALLFAILURE, "System call failure." },
		{ WSASYSNOTREADY, "Network subsystem is unavailable." },
		{ WSATRY_AGAIN, "Non-authoritative host not found." },
		{ WSAVERNOTSUPPORTED, "WINSOCK.DLL version out of range." }
	};

	socket::socket(blocking_type aBlocking, protocol_type aProtocol) : 
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(INVALID_SOCKET), iParent(0), iError(0),
		iAddress(INADDR_NONE), iPort(0), iConnected(false),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer)
	{
	}

	socket::socket(blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, bool aConnected) : 
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(aHandle), iParent(0), iError(0),
		iAddress(INADDR_NONE), iPort(0), iConnected(aConnected),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer)
	{
	}

	socket::socket(blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, u_long aAddress, u_short aPort, bool aConnected) : 
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(aHandle), iParent(0), iError(0),
		iAddress(aAddress), iPort(aPort), iConnected(aConnected),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer)
	{
	}

	socket::socket(blocking_type aBlocking, protocol_type aProtocol, u_long aAddress, u_short aPort, bool aConnect) : 
		iSocketIsBlocking(true), iBlocking(aBlocking), iProtocol(aProtocol), iHandle(INVALID_SOCKET), iParent(0), iError(0),
		iAddress(aAddress), iPort(aPort), iConnected(false),
		iInputBufferRecvPtr(iInputBuffer), iInputBufferNextPacketPtr(iInputBuffer), 
		iOutputBufferNextPacketPtr(iOutputBuffer)
	{
		if (aConnect)
			connect();
	}

	socket::~socket()
	{
		if (iHandle != INVALID_SOCKET)
			close();
	}

	bool socket::create(bool aCoalesce)
	{
		if (iHandle != INVALID_SOCKET)
			close();

		if ((iHandle = ::socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
		{
			get_error();
			return false;
		}
		
		if (!aCoalesce)
		{
			int optval = 1;
			if (setsockopt(iHandle, IPPROTO_TCP, TCP_NODELAY, 
				reinterpret_cast<char*>(&optval), sizeof(optval)) == SOCKET_ERROR)
			{
				get_error();
				close();
				return false;
			}
		}

		if (!set_blocking())
		{
			get_error();
			close();
			return false;
		}

		return true;
	}

	bool socket::connect(bool aCoalesce)
	{
		if (iConnected)
			close();

		if (iHandle == INVALID_SOCKET && !create(aCoalesce))
			return false;

		sockaddr_in sa;
		sa.sin_family = AF_INET;
		sa.sin_addr.s_addr = htonl(iAddress);
		sa.sin_port = htons(iPort);

		if (::connect(iHandle, reinterpret_cast<sockaddr*>(&sa), sizeof(sa)))
		{
			if (::WSAGetLastError() != WSAEWOULDBLOCK)
			{
				get_error();
				closesocket(iHandle);
				iHandle = INVALID_SOCKET;
				return false;
			}
		}
		else
			iConnected = true;

		return true;
	}

	void socket::close()
	{
		if (iHandle == INVALID_SOCKET)
			return;

		closesocket(iHandle);
		iHandle = INVALID_SOCKET;
		iConnected = false;
	}

	u_long socket::local_address() const
	{
		sockaddr_in sa; 
		int size = sizeof(sa);
		if (getsockname(iHandle, reinterpret_cast<sockaddr*>(&sa), &size) == INVALID_SOCKET)
			return 0;
		return ntohl(sa.sin_addr.s_addr);
	}

	bool socket::set_blocking(blocking_type aBlocking)
	{
		iBlocking = aBlocking;

		if (iBlocking == Blocking && iSocketIsBlocking)
			return true;

		if (iBlocking == NonBlocking && !iSocketIsBlocking)
			return true;

		u_long arg = (iBlocking == Blocking ? 0 : 1);
		if (ioctlsocket(iHandle, FIONBIO, &arg) == 0)
		{
			iSocketIsBlocking = (arg == 0 ? true : false);
			return true;
		}
		else
		{
			iBlocking = (iSocketIsBlocking ? Blocking : NonBlocking);
			get_error();
			return false;
		}
	}

	void socket::set_protocol(protocol_type aProtocol) 
	{ 
		if (iProtocol == aProtocol)
			return;
		
		iProtocol = aProtocol;
	}

	bool socket::has_input() const
	{ 
		return !iInputPackets.empty() || iInputBufferRecvPtr > iInputBuffer; 
	}
	
	bool socket::has_output() const
	{ 
		return !iOutputPackets.empty() || iOutputBufferNextPacketPtr > iOutputBuffer; 
	}

	void socket::get_error()
	{
		if (iHandle != INVALID_SOCKET)
		{
			int error;
			int optlen = sizeof(error);
			getsockopt(iHandle, SOL_SOCKET, SO_ERROR, reinterpret_cast<char*>(&error), &optlen);
			set_error(error);
		}
	}

	void socket::set_error(int aError)
	{
		iError = aError;
		iErrorString = "";
		for (size_t i = 0; i < sizeof(sSocketErrors) / sizeof(sSocketErrors[0]); ++i)
		{
			if (sSocketErrors[i].iErrorCode == iError)
			{
				iErrorString = sSocketErrors[i].iErrorString;
				break;
			}
		}
	}
	
	void socket::set_connected()
	{ 
		iConnected = true;
	}
	 
	 bool socket::handle_input(socket_manager& aManager)
	{
		int br = recv(iHandle, iInputBufferRecvPtr, (iInputBuffer + SocketBufferLength) - iInputBufferRecvPtr, 0);
		if (br == 0)
		{
			close();
			return true;
		}
		else if (br < 0) 
		{
			get_error();
			return false;
		}

		if (iProtocol == Packet)
		{
			size_t bytesLeft = (iInputBufferRecvPtr + br) - iInputBufferNextPacketPtr;
			while(bytesLeft >= PacketHeaderLength)
			{
				int id = *((int*) iInputBufferNextPacketPtr);
				size_t length = *reinterpret_cast<size_t*>(iInputBufferNextPacketPtr + sizeof(int));
				if (bytesLeft - PacketHeaderLength >= length)
				{
					if (id == 0 && length == 0) // combine split packet
					{
						bool first = true;
						packet_list::iterator largePacket;
						char* largePacketData;
						for (packet_list::iterator i = iInputPackets.begin(); i != iInputPackets.end();)
						{
							if (i->id() == 0)
							{
								if (first)
								{
									first = false;
									largePacket = iInputPackets.insert(i, packet(0, 0, 0));
									largePacket->set(*reinterpret_cast<int*>(i->data()), *reinterpret_cast<size_t*>(static_cast<char*>(i->data()) + sizeof(int)), 0);
									largePacketData = static_cast<char*>(largePacket->data());
									memcpy(largePacketData, static_cast<const char*>(i->data()) + PacketHeaderLength, i->length() - PacketHeaderLength);
									largePacketData += i->length() - PacketHeaderLength;
								}
								else
								{
									memcpy(largePacketData, i->data(), i->length());
									largePacketData += i->length();
								}
								iInputPackets.erase(i++);
							}
							else
								++i;
						}
					}
					else
					{
						iInputPackets.push_back(packet());
						iInputPackets.back().set(id, length, iInputBufferNextPacketPtr + PacketHeaderLength);
					}
					iInputBufferNextPacketPtr += length + PacketHeaderLength;
					bytesLeft = (iInputBufferRecvPtr + br) - iInputBufferNextPacketPtr;
				}
				else
					break;
			}
			memmove(iInputBuffer, iInputBufferNextPacketPtr, bytesLeft);
			iInputBufferNextPacketPtr = iInputBuffer;
			iInputBufferRecvPtr = iInputBuffer + bytesLeft;

			for (packet_list::iterator i = input().begin(); i != input().end();)
			{
				packet& thePacket = *i;
				if (thePacket.id() != 0)
				{
					aManager.notify_observers(socket_manager_observer::NotifyData, *this, thePacket);
					iInputPackets.erase(i++);
				}
				else
					++i;
			}
		}
		else
		{
			iBasicProtocolPacket.set(0, br, iInputBufferRecvPtr);
			aManager.notify_observers(socket_manager_observer::NotifyData, *this, iBasicProtocolPacket);
		}

		return true;
	}

	bool socket::handle_output(socket_manager& aManager)
	{
		size_t bytesLeft = (iOutputBuffer + SocketBufferLength) - iOutputBufferNextPacketPtr;
		size_t headerSize = (iProtocol == Packet ? PacketHeaderLength : 0); 
		while(bytesLeft >= headerSize && !iOutputPackets.empty())
		{
			if (iOutputPackets.front().length() > MaxPacketLength) // split large packets up
			{
				packet_list::iterator largePacket = iOutputPackets.begin();
				char headerPacket[MaxPacketLength];
				*reinterpret_cast<int*>(headerPacket) = largePacket->id();
				*reinterpret_cast<size_t*>(headerPacket + sizeof(int)) = largePacket->length();
				memcpy(headerPacket + headerSize, largePacket->data(), MaxPacketLength - headerSize);
				iOutputPackets.insert(largePacket, packet(0, MaxPacketLength, headerPacket));
				size_t largePacketBytesLeft = largePacket->length() - (MaxPacketLength - headerSize);
				const char* largePacketData = static_cast<const char*>(largePacket->data()) + (MaxPacketLength - headerSize);
				while (largePacketBytesLeft > 0)
				{
					size_t nextSize = largePacketBytesLeft > MaxPacketLength ? MaxPacketLength : largePacketBytesLeft;
					iOutputPackets.insert(largePacket, packet(0, nextSize, largePacketData));
					largePacketData += nextSize;
					largePacketBytesLeft -= nextSize;
				}
				iOutputPackets.insert(largePacket, packet(0, 0, 0));
				iOutputPackets.erase(largePacket);
			}

			if (iProtocol == Packet)
			{
				int id = iOutputPackets.front().id();
				size_t length = iOutputPackets.front().length();
				if (bytesLeft - PacketHeaderLength >= length)
				{
					*((int*) iOutputBufferNextPacketPtr) =  id;
					*((int*) iOutputBufferNextPacketPtr + 1) = length;
					if (length > 0)
						memcpy(iOutputBufferNextPacketPtr + PacketHeaderLength, iOutputPackets.front().data(), length);
					iOutputPackets.pop_front();
					iOutputBufferNextPacketPtr += length + PacketHeaderLength;
					bytesLeft = (iOutputBuffer + SocketBufferLength) - iOutputBufferNextPacketPtr;
				}
				else
					break;
			}
			else // iProtocol == Basic
			{
				size_t length = iOutputPackets.front().length();
				if (bytesLeft >= length)
				{
					if (length > 0)
						memcpy(iOutputBufferNextPacketPtr, iOutputPackets.front().data(), length);
					iOutputPackets.pop_front();
					iOutputBufferNextPacketPtr += length;
					bytesLeft = (iOutputBuffer + SocketBufferLength) - iOutputBufferNextPacketPtr;
				}
				else
					break;
			}
		}
		
		int bs = send(iHandle, iOutputBuffer, iOutputBufferNextPacketPtr - iOutputBuffer, 0);
		if (bs == SOCKET_ERROR)
		{
			get_error();
			return false;
		}

		bytesLeft = (iOutputBufferNextPacketPtr - iOutputBuffer) - bs;
		if (bytesLeft > 0)
			memmove(iOutputBuffer, iOutputBuffer + bs, bytesLeft);
		iOutputBufferNextPacketPtr -= bs;

		return true;
	}

	bool socket::handle_error(socket_manager& aManager)
	{
		get_error();
		return false;
	}

	
	listen_socket::listen_socket(blocking_type aBlocking, protocol_type aProtocol, u_short aPort) : socket(aBlocking, aProtocol, 0, aPort, false)
	{
		listen();
	}

	bool listen_socket::handle_input(socket_manager& aManager)
	{
		SOCKET s;
		sockaddr_in sa; int size = sizeof(sa);
		if ((s = accept(iHandle, reinterpret_cast<sockaddr*>(&sa), &size)) == INVALID_SOCKET)
		{
			get_error();
			return false;
		}
		if (getpeername(s, reinterpret_cast<sockaddr*>(&sa), &size) == INVALID_SOCKET)
		{
			get_error();
			closesocket(s);
			return false;
		}
		int opt = 1;
		setsockopt(s, SOL_SOCKET, SO_DONTLINGER, reinterpret_cast<char*>(opt), sizeof(opt));

		socket* newSocket = new socket(iBlocking, iProtocol, s, ntohl(sa.sin_addr.s_addr), ntohs(sa.sin_port), true);
		newSocket->set_parent(this);
		aManager.add_socket(newSocket);

		return true;
	}

	bool listen_socket::listen()
	{
		if (iHandle != INVALID_SOCKET)
			close();
		
		if ((iHandle = ::socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
			return false;

		int optval = 1;
		if (setsockopt(iHandle, SOL_SOCKET, SO_REUSEADDR, 
			reinterpret_cast<char*>(&optval), sizeof(optval)) == SOCKET_ERROR)
		{
			get_error();
			close();
			return false;
		}

		sockaddr_in sa;
		memset(&sa, 0, sizeof(sa));
		sa.sin_family = AF_INET;
		sa.sin_port = htons(iPort);

		if (bind(iHandle, reinterpret_cast<sockaddr*>(&sa), sizeof(sa)) == SOCKET_ERROR)
		{
			get_error();
			close();
			return false;
		}

		if (!set_blocking())
		{
			get_error();
			close();
			return false;
		}

		if (::listen(iHandle, 256) == SOCKET_ERROR)
		{
			get_error();
			close();
			return false;
		}

		iConnected = true;

		return true;
	}


	resolver::resolver()
	{
		::InitializeCriticalSection(&iCriticalSection);
	}

	resolver::~resolver()
	{
		lock();
		for (entries::iterator i = iEntries.begin(); i != iEntries.end(); ++i)
		{
			if (!i->iReady)
				::TerminateThread(i->iThreadHandle, 0);
			::CloseHandle(i->iThreadHandle);
		}
		unlock();
		::DeleteCriticalSection(&iCriticalSection);
	}

	void resolver::add(const socket* aSocket, const std::string& aHostName)
	{
		lock();
		iEntries.push_back(entry());
		iEntries.back().iSocket = aSocket;
		iEntries.back().iHostName = aHostName;
		iEntries.back().iThreadHandle = ::CreateThread(NULL, 0, request, this, CREATE_SUSPENDED, &iEntries.back().iThreadId);
		::ResumeThread(iEntries.back().iThreadHandle);
		unlock();
	}

	void resolver::remove(const socket* aSocket)
	{
		lock();
		for (entries::iterator i = iEntries.begin(); i != iEntries.end();)
		{
			if (i->iSocket == aSocket)
			{
				::TerminateThread(i->iThreadHandle, 0);
				::CloseHandle(i->iThreadHandle);
				iEntries.erase(i++);
			}
			else
				++i;
		}
		unlock();
	}

	bool resolver::next_ready(entry& aEntry)
	{
		lock();
		bool found = false;
		for (entries::iterator i = iEntries.begin(); !found && i != iEntries.end(); ++i)
		{
			if (i->iReady)
			{
				aEntry = *i;
				::CloseHandle(i->iThreadHandle);
				iEntries.erase(i);
				found = true;
			}
		}
		unlock();
		return found;
	}

	resolver::entries::iterator resolver::find(DWORD aThreadId)
	{
		for (entries::iterator i = iEntries.begin(); i != iEntries.end(); ++i)
			if (i->iThreadId == aThreadId)
				return i;
		return iEntries.end();
	}

	void resolver::lock()
	{
		::EnterCriticalSection(&iCriticalSection);
	}

	void resolver::unlock()
	{
		::LeaveCriticalSection(&iCriticalSection);
	}

	DWORD WINAPI resolver::request(LPVOID aResolver)
	{
		resolver& theResolver = *static_cast<resolver*>(aResolver);
		std::string hostName;
		theResolver.lock();
		entries::iterator i = theResolver.find(::GetCurrentThreadId());
		if (i != theResolver.iEntries.end())
		{
			// we are multi-threaded so we don't want these std::strings to
			// share a pointer so use assign instead of =
			hostName.assign(i->iHostName.begin(), i->iHostName.end());
		}
		theResolver.unlock();
		int error = 0;
		u_long address = INADDR_NONE;
		if (!hostName.empty())
		{
			hostent *he = gethostbyname(hostName.c_str());
			if (he == NULL)
				error = ::WSAGetLastError();
			else
			{
				sockaddr_in sa;
				memcpy(&sa.sin_addr, he->h_addr, he->h_length);
				address = ntohl(sa.sin_addr.s_addr);
			}
		}
		theResolver.lock();
		i = theResolver.find(::GetCurrentThreadId());
		if (i != theResolver.iEntries.end())
		{
			i->iAddress = address;
			i->iError = error;
			i->iReady = true;
		}
		theResolver.unlock();
		return 0;
	}

	
	socket_manager::socket_manager(socket::blocking_type aBlocking, socket::protocol_type aProtocol) : iBlocking(aBlocking), iProtocol(aProtocol), iError(0), iInitialised(false), iInSelect(false)
	{
		WSADATA wsaData;
		if (::WSAStartup(MAKEWORD(2, 0), &wsaData) == 0)
			iInitialised = true;
		else
			get_error();
	}

	socket_manager::~socket_manager()
	{
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end(); ++i)
		{
			delete *i;
		}
		iSockets.erase(iSockets.begin(), iSockets.end());

		if (iInitialised)
			::WSACleanup();
	}

	socket* socket_manager::listen(u_short aPort)
	{
		listen_socket* theSocket = new listen_socket(iBlocking, iProtocol, aPort);
		if (!theSocket->listen())
		{
			get_error();
			delete theSocket;
			return 0;
		}
	
		add_socket(theSocket, true);
		return theSocket;
	}

	socket* socket_manager::connect(u_long aAddress, u_short aPort)
	{
		socket* theSocket = new socket(iBlocking, iProtocol, aAddress, aPort, false);
		if (!theSocket->connect())
		{
			get_error();
			delete theSocket;
			theSocket = 0;
		}
		else
		{
			add_socket(theSocket);
			if (theSocket->connected())
				notify_observers(socket_manager_observer::NotifySocketConnected, *theSocket);
		}
		return theSocket;
	}

	socket* socket_manager::connect(const std::string& aHostName, u_short aPort)
	{
		socket* theSocket = 0;

		u_long address = to_address(aHostName);

		if (address != INADDR_NONE)
			theSocket = connect(address, aPort);
		else if (iBlocking == socket::NonBlocking)
		{
			theSocket = new socket(iBlocking, iProtocol, static_cast<u_long>(INADDR_NONE), aPort, false);
			if (!theSocket->create())
			{
				get_error();
				delete theSocket;
				theSocket = 0;
			}
			else
			{
				add_socket(theSocket);
				iResolver.add(theSocket, aHostName);
			}
		}

		return theSocket;
	}

	bool socket_manager::select()
	{
		if (iInSelect)
			return false;

		iInSelect = true;
		exit_value<bool> ev(iInSelect, false);

		// remove zombies
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			if ((*i)->zombie())
				remove_socket(*i++);
			else
				++i;
		}

		resolver::entry resolverEntry;
		while(iResolver.next_ready(resolverEntry))
		{
			socket* theSocket = const_cast<socket*>(resolverEntry.iSocket);
			theSocket->iAddress = resolverEntry.iAddress;
			if (theSocket->address() != INADDR_NONE)
			{
				if (!theSocket->connect())
				{
					get_error();
					remove_socket(theSocket);
				}
				else
				{
					if (theSocket->connected())
						notify_observers(socket_manager_observer::NotifySocketConnected, *theSocket);
				}
			}
			else
			{
				theSocket->set_error(resolverEntry.iError);
				remove_socket(theSocket);
			}
		}

		if (iSockets.empty())
			return true;

		fd_set readfds;
		fd_set writefds;
		fd_set exceptfds;
		
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		FD_ZERO(&exceptfds);
		size_t count = 0;
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			socket& theSocket = **i++;

			if (theSocket.address() == INADDR_NONE)
				continue;

			SOCKET handle = theSocket.handle();
			FD_SET(handle, &readfds);
			if (theSocket.has_output() || !theSocket.connected())
				FD_SET(handle, &writefds);
			FD_SET(handle, &exceptfds);

			if (++count == FD_SETSIZE || i == iSockets.end())
			{
				if (!do_select(readfds, writefds, exceptfds))
					return false;
				FD_ZERO(&readfds);
				FD_ZERO(&writefds);
				FD_ZERO(&exceptfds);
				count = 0;
			}
		}
		return true;
	}

	bool socket_manager::connected() const
	{
		for (socket_list::const_iterator i = iSockets.begin(); i != iSockets.end(); ++i)
			if ((*i)->connected())
				return true;
		return false;
	}

	void socket_manager::disconnect()
	{
		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end(); ++i)
			(*i)->close();
	}

	void socket_manager::disconnect(socket& aSocket)
	{
		aSocket.close();
	}

	class next_socket
	{
	public:
		next_socket(socket_manager& aSocketManager, bool& aError, bool& aClosed, socket_manager::socket_list::iterator& aIterator) : 
			iSocketManager(aSocketManager), iError(aError), iClosed(aClosed), iIterator(aIterator) {}
		~next_socket()
		{
			if (!iError && !iClosed)
				++iIterator;
			else
				iSocketManager.remove_socket(*iIterator++);
		}
	private:
		socket_manager& iSocketManager;
		bool& iError;
		bool& iClosed;
		socket_manager::socket_list::iterator& iIterator;
	};

	bool socket_manager::do_select(fd_set& aReadFds, fd_set& aWriteFds, fd_set& aExceptFds)
	{
		timeval timeout = {0, 0 };
		if (::select(0, &aReadFds, &aWriteFds, &aExceptFds, &timeout) == SOCKET_ERROR)
		{
			get_error();
			return false;
		}

		for (socket_list::iterator i = iSockets.begin(); i != iSockets.end();)
		{
			socket& theSocket = **i;
			if (theSocket.zombie())
			{
				++i;
				continue;
			}

			bool error = false;
			bool closed = false;
			next_socket ns(*this, error, closed, i);

			SOCKET handle = theSocket.handle();

			if (FD_ISSET(handle, &aWriteFds))
			{
				if (!(*i)->connected())
				{
					(*i)->set_connected();
					notify_observers(socket_manager_observer::NotifySocketConnected, theSocket);
				}
				else
					if (error = !(*i)->handle_output(*this))
						continue;
			}

			if (FD_ISSET(handle, &aReadFds))
				if ((error = !(*i)->handle_input(*this)) || (closed = theSocket.zombie()))
					continue;

			if (FD_ISSET(handle, &aExceptFds))
				if(error = !(*i)->handle_error(*this))
					continue;
		}
		return true;
	}

	void socket_manager::get_error()
	{
		iError = ::WSAGetLastError();
		iErrorString = "";
		for (size_t i = 0; i < sizeof(sSocketErrors) / sizeof(sSocketErrors[0]); ++i)
		{
			if (sSocketErrors[i].iErrorCode == iError)
			{
				iErrorString = sSocketErrors[i].iErrorString;
				break;
			}
		}
	}

	u_long socket_manager::to_address(const std::string& aHostName)
	{
		u_long address = inet_addr(aHostName.c_str());
		if (address != INADDR_NONE)
			return ntohl(address);

		if (iBlocking == socket::NonBlocking)
			return INADDR_NONE;

		hostent *he = gethostbyname(aHostName.c_str());
		if (he == NULL)
		{
			get_error();
			return INADDR_NONE;
		}
		
		sockaddr_in sa;
		memcpy(&sa.sin_addr, he->h_addr, he->h_length);
		return ntohl(sa.sin_addr.s_addr);
	}

	void socket_manager::add_socket(socket* aSocket, bool aListener)
	{
		iSockets.push_back(aSocket);
		notify_observers(socket_manager_observer::NotifySocketAdded, *aSocket, aListener);
		if (aSocket->connected())
			notify_observers(socket_manager_observer::NotifySocketConnected, *aSocket);
	}

	void socket_manager::remove_socket(socket* aSocket)
	{
		iSockets.remove(aSocket);
		iResolver.remove(aSocket);
		notify_observers(socket_manager_observer::NotifySocketRemoved, *aSocket);
		delete aSocket;
	}

	void socket_manager::notify_observer(socket_manager_observer& aObserver, socket_manager_observer::notify_type aType, const void* aParameter, const void* aParameter2)
	{
		socket& theSocket = *const_cast<socket*>(reinterpret_cast<const socket*>(aParameter));
		switch(aType)
		{
		case socket_manager_observer::NotifySocketAdded:
			aObserver.socket_added(theSocket, *reinterpret_cast<const bool*>(aParameter2));
			break;
		case socket_manager_observer::NotifySocketRemoved:
			aObserver.socket_removed(theSocket);
			break;
		case socket_manager_observer::NotifySocketConnected:
			aObserver.socket_connected(theSocket);
			break;
		case socket_manager_observer::NotifyData:
			aObserver.socket_data(theSocket, *reinterpret_cast<const packet*>(aParameter2));
			break;
		}
	}
} // namespace lib