// network.h - v1.0
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_NETWORK
#define LIB_NETWORK

#include <winsock2.h>
#include <list>
#include "observable.h"

namespace lib
{
	enum { MaxPacketLength = 512 };
	enum { PacketHeaderLength = sizeof(int) + sizeof(size_t) };
	enum { SocketBufferLength = MaxPacketLength + PacketHeaderLength };

	class packet
	{
	public:
		// construction
		packet();
		packet(const packet& aPacket);
		packet(int aId);
		packet(int aId, size_t aLength, const void* aData);
		packet(int aId, char aData);
		packet(int aId, int aData);
		packet(size_t aLength, const void* aData);
		~packet();
	public:
		// operations
		void set(int aId, size_t aLength, const void* aData);
		void set(int aId, char aData);
		void set(int aId, int aData);
		void set(size_t aLength, const void* aData);
		int id() const { return iId; }
		size_t length() const { return iLength; }
		const void* data() const { return iData; }
		void* data() { return iData; }
	private:
		// attributes
		int iId;
		size_t iLength;
		size_t iAllocLength;
		void* iData;
	};

	// WARNING: Think about endianness if using packet_value<T>
	template <typename T>
	T packet_value(const packet& aPacket)
	{
		assert(aPacket.length() == sizeof(T));
		return *reinterpret_cast<const T*>(aPacket.data());
	}
	template <>
	std::string packet_value<std::string>(const packet& aPacket);

	class serializable;
	void save_to_packet(packet& aPacket, serializable& aObject);
	void load_from_packet(const packet& aPacket, serializable& aObject);

	typedef std::list<packet> packet_list;

	class socket
	{
		friend class socket_manager; 
	public:
		// types
		enum blocking_type { Blocking, NonBlocking };
		enum protocol_type { Packet, Basic };
	public:
		// construction
		socket(blocking_type aBlocking, protocol_type aProtocol);
		socket(blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, bool aConnected = false);
		socket(blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, u_long aAddress, u_short aPort, bool aConnected = false);
		socket(blocking_type aBlocking, protocol_type aProtocol, u_long aAddress, u_short aPort, bool aConnect = true);
		virtual ~socket();
	public:
		// operations
		bool create(bool aCoalesce = false);
		bool connect(bool aCoalesce = false);
		bool connected() const { return iConnected; }
		void close();
		SOCKET handle() const { return iHandle; }
		u_long address() const { return iAddress; }
		u_long local_address() const;
		bool set_blocking(blocking_type aBlocking);
		bool set_blocking() { return set_blocking(iBlocking); }
		blocking_type blocking() const { return iBlocking; }
		void set_protocol(protocol_type aProtocol);
		protocol_type protocol() const { return iProtocol; }
		packet_list& input() { return iInputPackets; }
		packet_list& output() { return iOutputPackets; }
		bool has_input() const;
		bool has_output() const;
		u_short port() const { return iPort; }
		socket* parent() const { return iParent; }
		void set_parent(socket* aParent) { iParent = aParent; }
		bool zombie() const { return iHandle == INVALID_SOCKET; }
		int error() const { return iError; }
		const std::string& error_string() const { return iErrorString; }
	private:
		// implementation
		void set_connected();
		virtual bool handle_input(socket_manager& aManager);
		virtual bool handle_output(socket_manager& aManager);
		virtual bool handle_error(socket_manager& aManager);
	protected:
		// implementation
		void get_error();
		void set_error(int aError);
	protected:
		// attributes
		bool iSocketIsBlocking;
		blocking_type iBlocking;
		protocol_type iProtocol;
		SOCKET iHandle;
		socket* iParent;
		int iError;
		std::string iErrorString;
		u_long iAddress;
		u_short iPort;
		bool iConnected;
		char iInputBuffer[SocketBufferLength];
		char* iInputBufferRecvPtr;
		char* iInputBufferNextPacketPtr;
		char iOutputBuffer[SocketBufferLength];
		char* iOutputBufferNextPacketPtr;
		packet_list iInputPackets;
		packet_list iOutputPackets;
		packet iBasicProtocolPacket;
	};

	class listen_socket : public socket
	{
	public:
		// construction
		listen_socket(blocking_type aBlocking, protocol_type aProtocol, u_short aPort);
	public:
		// operations
		bool listen();
	private:
		// implementation
		virtual bool handle_input(socket_manager& aManager);
		virtual bool handle_output(socket_manager& aManager) { return true; }
	};

	class resolver
	{
	public:
		// types
		struct entry
		{
			const socket* iSocket;
			std::string iHostName;
			u_long iAddress;
			HANDLE iThreadHandle;
			DWORD iThreadId;
			bool iReady;
			int iError;
			entry() : iSocket(0), iAddress(INADDR_NONE), iThreadId(0), iReady(false), iError(0) {}
		};
		typedef std::list<entry> entries;
	public:
		// construction
		resolver();
		~resolver();
	public:
		// operations
		void add(const socket* aSocket, const std::string& aHostName);
		void remove(const socket* aSocket);
		bool next_ready(entry& aEntry);
	private:
		// implementation
		entries::iterator find(DWORD aThreadId);
		void lock();
		void unlock();
		static DWORD WINAPI request(LPVOID aResolver);
	private:
		// attributes
		entries iEntries;
		CRITICAL_SECTION iCriticalSection;
	};

	class socket_manager_observer
	{
		friend class socket_manager;
	private:
		virtual void socket_added(socket& aSocket, bool aListener) = 0;
		virtual void socket_removed(socket& aSocket) = 0;
		virtual void socket_connected(socket& aSocket) = 0;
		virtual void socket_data(socket& aSocket, const packet& aPacket) = 0;
	public:
		enum notify_type { NotifySocketAdded, NotifySocketRemoved, NotifySocketConnected, NotifyData };
	};

	class socket_manager : public observable<socket_manager_observer>
	{
		friend class socket;
		friend class listen_socket;
		friend class next_socket;
	public:
		// construction
		socket_manager(socket::blocking_type aBlocking = socket::NonBlocking, socket::protocol_type aProtcol = socket::Packet);
		~socket_manager();
	public:
		// operations
		socket* listen(u_short aPort);
		socket* connect(u_long aAddress, u_short aPort);
		socket* connect(const std::string& aHostName, u_short aPort);
		bool empty() const { return iSockets.empty(); }
		bool select();
		bool connected() const;
		void disconnect();
		void disconnect(socket& aSocket);
		int error() { get_error(); return iError; }
		const std::string& error_string() { get_error(); return iErrorString; }
	private:
		// implementation
		bool do_select(fd_set& aReadFds, fd_set& aWriteFds, fd_set& aExceptFds);
		void get_error();
		u_long to_address(const std::string& aHostName);
		void add_socket(socket* aSocket, bool aListener = false);
		void remove_socket(socket* aSocket);
		// from observable<socket_manager_observer>
		virtual void notify_observer(socket_manager_observer& aObserver, socket_manager_observer::notify_type aType, const void* aParameter = 0, const void* aParameter2 = 0);
	private:
		// attributes
		socket::protocol_type iProtocol;
		socket::blocking_type iBlocking;
		int iError;
		std::string iErrorString;
		bool iInitialised;
		bool iInSelect;
		resolver iResolver;
		typedef std::list<socket*> socket_list;
		socket_list iSockets;
	};
}

#endif // LIB_NETWORK