// serialization.h - v1.0.1
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#if !defined(LIB_SERIALIZATION)
#define LIB_SERIALIZATION

#include <climits>

// WARNING: This serialization library does not yet support endianness

namespace lib
{
	typedef unsigned char byte_t;
	typedef unsigned int dword_t;

	class instream
	{
	public:
		virtual byte_t read_char() = 0;
		virtual dword_t read_int() = 0;
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void read_string(std::string& aString) = 0;
#endif
	};
	class outstream
	{
	public:
		virtual void write_char(byte_t aChar) = 0;
		virtual void write_int(dword_t aInt) = 0;
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void write_string(const std::string& aString) = 0;
#endif
	};
	class serializable
	{
	public:
		virtual void load(instream& aLoader) = 0;
		virtual void save(outstream& aSaver) const = 0;
	};

	class ptr_instream : public instream
	{
	public:
		struct error {};
		ptr_instream(const void* aPtr, size_t aLength) : iPtr(aPtr), iLength(aLength) {}
	private:
		virtual byte_t read_char()
		{
			if (iLength < sizeof(byte_t))
				throw error();
			byte_t ret = *static_cast<const byte_t*>(iPtr);
			iPtr = static_cast<const byte_t*>(iPtr) + 1;
			iLength -= sizeof(byte_t);
			return ret;
		}
		virtual dword_t read_int()
		{
			if (iLength < sizeof(dword_t))
				throw error();
			dword_t ret = 0;
			for (size_t i = 0; i < sizeof(dword_t); ++i)
				ret = (ret << CHAR_BIT) | read_char();
			return ret;
		}
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void read_string(std::string& aString)
		{
			size_t length = read_int();
			if (iLength < length)
				throw error();
			aString.assign(static_cast<const char*>(iPtr), length);
			iPtr = static_cast<const char*>(iPtr) + length;
			iLength -= length;
		}
#endif
	private:
		const void* iPtr;
		size_t iLength;
	};
	class ptr_outstream : public outstream
	{
	public:
		struct error {};
		ptr_outstream(void* aPtr, size_t aLength) : iPtr(aPtr), iLength(aLength) {}
	private:
		virtual void write_char(byte_t aChar)
		{
			if (iLength < sizeof(byte_t))
				throw error();
			*static_cast<byte_t*>(iPtr) = aChar;
			iPtr = static_cast<byte_t*>(iPtr) + 1;
			iLength -= sizeof(byte_t);
		}
		virtual void write_int(dword_t aInt)
		{
			if (iLength < sizeof(dword_t))
				throw error();
			byte_t bytes[sizeof(dword_t)];
			for (size_t i = 0; i < sizeof(dword_t); ++i)
			{
				bytes[(sizeof(dword_t) - i) - 1] = aInt & UCHAR_MAX;
				aInt >>= CHAR_BIT;
			}
			for (size_t i = 0; i < sizeof(dword_t); ++i)
				write_char(bytes[i]);
		}
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void write_string(const std::string& aString)
		{
			write_int(aString.length());
			if (iLength < aString.length())
				throw error();
			memcpy(iPtr, aString.data(), aString.length());
			iPtr = static_cast<byte_t*>(iPtr) + aString.length();
			iLength -= aString.length();
		}
#endif
	private:
		void* iPtr;
		size_t iLength;
	};
	class len_outstream : public outstream
	{
	public:
		len_outstream(size_t& aLength) : iLength(aLength) { iLength = 0; }
	private:
		virtual void write_char(byte_t /*aChar*/)
		{
			iLength += sizeof(byte_t);
		}
		virtual void write_int(dword_t /*aInt*/)
		{
			iLength += sizeof(dword_t);
		}
#ifdef LIB_SERIALIZATION_USE_STRING
		virtual void write_string(const std::string& aString)
		{
			write_int(aString.length());
			iLength += aString.length();
		}
#endif
	private:
		size_t& iLength;
	};
}

#endif // LIB_SERIALIZATION