// smart_ptr.h - v1.2.3
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

/* WARNING: This class is not a substitute for the shared_ptr available on your platform
 * which you should be using instead.  It is here either for technical reasons or for when 
 * there is no standard library available.
  */

#ifndef LIB_SMART_PTR
#define LIB_SMART_PTR

namespace lib
{
	template <typename T>
	class smart_ptr
	{
	public:
		typedef T value_type;
		typedef T* pointer;
		typedef const T* const_pointer;
		typedef T& reference;
		typedef const T& const_reference;

	private:
		T* iPointer;
		typedef size_t counter_type;
		struct counter
		{
		public:
			counter() : iCounter(new counter_type) 
			{
				*iCounter = 0;
			}
			counter(counter_type* aCounter) : iCounter(aCounter) {}
			counter(const counter& aOther) : iCounter(aOther.iCounter) {}
			~counter() 
			{ 
				if (iCounter != 0 && *iCounter == 0) 
					delete iCounter; 
			}
			counter& operator=(const counter& aOther) 
			{ 
				iCounter = aOther.iCounter; 
				return *this;
			}
			counter_type operator++() 
			{ 
				if (iCounter != 0)
				{
					++*iCounter; 
					return *iCounter; 
				}
				else
					return 0;
			}
			counter_type operator--() 
			{ 
				if (iCounter != 0)
				{
					--*iCounter;
					return *iCounter;
				}
				else
					return 0;
			}
		private:
			counter_type* iCounter;
		} iReferenceCount;
	
	public:
		// construction
		smart_ptr() : iPointer(0), iReferenceCount(0) {}
		smart_ptr(pointer aPointer) : iPointer(aPointer), iReferenceCount()
		{
			++iReferenceCount;
		}
		smart_ptr(const smart_ptr& aOther) : iPointer(aOther.iPointer), iReferenceCount(aOther.iReferenceCount)
		{
			++iReferenceCount;
		}
		~smart_ptr()
		{
			if (iPointer != 0)
				if (--iReferenceCount == 0)
					delete iPointer;
		}
		// element access
		reference operator*() { return *iPointer; }
		const_reference operator*() const { return *iPointer; }
		pointer operator->() { return iPointer; }
		const_pointer operator->() const { return iPointer; }
		pointer get() { return iPointer; }
		const_pointer get() const { return iPointer; }
		pointer release() 
		{ 
			pointer ret = iPointer;
			if (iPointer != 0)
			{
				iPointer = 0;
				--iReferenceCount;
			}
			return ret;
		}
		// modifiers
		void reset()
		{
			smart_ptr().swap(*this);
		}
		smart_ptr& operator=(pointer aPointer)
		{
			smart_ptr(aPointer).swap(*this);
			return *this;
		}
		smart_ptr& operator=(const smart_ptr& aOther)
		{
			smart_ptr(aOther).swap(*this);
			return *this;
		}
		void swap(smart_ptr& aOther)
		{
			std::swap(iPointer, aOther.iPointer);
			std::swap(iReferenceCount, aOther.iReferenceCount);
		}
		// comparison
		bool operator==(const smart_ptr& aOther) const
		{
			return get() == aOther.get();
		}
		bool operator!=(const smart_ptr& aOther) const
		{
			return get() != aOther.get();
		}
		bool operator<(const smart_ptr& aOther) const
		{
			return count() < aOther.count();
		}

	private:
		// implementation
		counter_type count() const
		{
			return iReferenceCount.count();
		}
	};

	template <typename T>
	inline smart_ptr<T> make_smart_ptr(T* aPointer)
	{
		return smart_ptr<T>(aPointer);
	}
}

#endif // LIB_SMART_PTR