// string.h - v1.0.3
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_STRING
#define LIB_STRING

#include <algorithm>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cwchar>
#include <cctype>
#include <cwctype>

namespace lib 
{
	template <typename C, typename FwdIter1, typename FwdIter2>
	inline FwdIter1 tokens(FwdIter1 aFirst, FwdIter1 aLast, FwdIter2 aDelimeterFirst, FwdIter2 aDelimiterLast, C& aTokens, std::size_t aMaxTokens = 0, bool aMergeDelimeters = true)
	{
		if (aFirst >= aLast)
			return aFirst;

		typedef typename C::value_type value_type;

		FwdIter1 b = aFirst;
		FwdIter1 e = std::find_first_of(b, aLast, aDelimeterFirst, aDelimiterLast);
		std::size_t tokens = 0;
		while(e != aLast && (aMaxTokens == 0 || tokens < aMaxTokens))
		{
			if (b == e && !aMergeDelimeters)
			{
				aTokens.push_back(value_type(b, b));
				++tokens;
			}
			else if (b != e)
			{
				aTokens.push_back(value_type(b, e));
				++tokens;
			}
			b = e;
			++b;
			e = std::find_first_of(b, aLast, aDelimeterFirst, aDelimiterLast);
		}
		if (b != e && (aMaxTokens == 0 || tokens < aMaxTokens))
		{
			aTokens.push_back(value_type(b, e));
			b = e;
		}
		return b;
	}
	template <typename CharT, typename C>
	inline void tokens(const std::basic_string<CharT>& aLine, const std::basic_string<CharT>& aDelimeter, C& aTokens, std::size_t aMaxTokens = 0, bool aMergeDelimeters = true)
	{
		tokens(aLine.begin(), aLine.end(), aDelimeter.begin(), aDelimeter.end(), aTokens, aMaxTokens, aMergeDelimeters);
	}

	template <typename CharT>
	inline void remove_trailing(std::basic_string<CharT>& aString, const std::basic_string<CharT>& aTrailing)
	{
		typename std::basic_string<CharT>::std::size_type pos = aString.find_last_not_of(aTrailing);
		if (pos != std::basic_string<CharT>::npos)
			aString.erase(aString.begin() + pos + 1, aString.end());
	}

	template <typename CharT>
	inline long string_to_integer(const std::basic_string<CharT>& aString, std::size_t aBase = 10)
	{
		return strtol(aString.c_str(), 0, aBase);
	}

	template <>
	inline long string_to_integer<wchar_t>(const std::basic_string<wchar_t>& aString, std::size_t aBase)
	{
		return wcstol(aString.c_str(), 0, aBase);
	}

	template <typename CharT>
	inline std::basic_string<CharT> integer_to_string(long aInteger, std::size_t aBase = 10, std::size_t aWidth = 0, CharT aFill = '0')
	{
		std::basic_stringstream<CharT> string;
		if (aBase == 16)
			string << std::hex << std::uppercase;
		if (aWidth != 0)
		{
			string.width(aWidth);
			string.fill(aFill);
		}
		string << aInteger;
		return string.str();
	}

	template <>
	inline std::basic_string<wchar_t> integer_to_string<wchar_t>(long aInteger, std::size_t aBase, std::size_t aWidth, wchar_t aFill)
	{
		std::basic_stringstream<wchar_t> string;
		if (aBase == 16)
			string << std::hex;
		if (aWidth != 0)
		{
			string.width(aWidth);
			string.fill(aFill);
		}
		string << aInteger;
		return string.str();
	}

	inline std::string wide_to_utf8(const std::wstring& aString)
	{
		// TODO
		std::string narrowString;
		for (std::wstring::const_iterator i = aString.begin(); i != aString.end(); ++i)
			narrowString.append(1, static_cast<char>(*i));
		return narrowString;
	}

	inline std::wstring utf8_to_wide(const std::string& aString)
	{
		// TODO
		std::wstring wideString;
		for (std::string::const_iterator i = aString.begin(); i != aString.end(); ++i)
			wideString.append(1, static_cast<wchar_t>(*i));
		return wideString;
	}

	template <typename CharT>
	inline CharT tolower(int c, CharT=0) {	return ::tolower(c); }
	template <>
	inline wchar_t tolower<wchar_t>(int c, wchar_t) { return ::towlower(c); }

	template <typename Traits>
	struct ci_char_traits : Traits
	{
	public:
		typedef typename Traits::char_type char_type;
		typedef typename Traits::int_type int_type;
	public:
		static int compare(const char_type* s1, const char_type* s2, std::size_t n)
		{
			for(;n-- > 0;++s1, ++s2)
			{
				if (eq(*s1, *s2))
					continue;
				else if (lt(*s1, *s2))
					return -1;
				else
					return 1;
			}
			return 0;
		}
		static bool eq(const char_type& c1, const char_type& c2)
		{
			return lower(c1) == lower(c2);
		}
		static bool lt(const char_type& c1, const char_type& c2)
		{
			return lower(c1) < lower(c2);
		}
		static int_type lower(char_type c)
		{
			return tolower<char_type>(to_int_type(c), char_type());
		}
	};

	typedef std::basic_string<char, ci_char_traits<std::char_traits<char> > > ci_string;
	typedef std::basic_string<wchar_t, ci_char_traits<std::char_traits<wchar_t> > > ci_wstring;

	inline ci_string make_ci_string(const std::string& s)
	{
		return ci_string(s.begin(), s.end());
	}
	inline ci_wstring make_ci_string(const std::wstring& s)
	{
		return ci_wstring(s.begin(), s.end());
	}

	inline bool operator==(const ci_string& s1, const std::string& s2)
	{
		return s1 == ci_string(s2.begin(), s2.end());
	}
	inline bool operator==(const std::string& s1, const ci_string& s2)
	{
		return ci_string(s1.begin(), s1.end()) == s2;
	}
	inline bool operator==(const ci_wstring& s1, const std::wstring& s2)
	{
		return s1 == ci_wstring(s2.begin(), s2.end());
	}
	inline bool operator==(const std::wstring& s1, const ci_wstring& s2)
	{
		return ci_wstring(s1.begin(), s1.end()) == s2;
	}
	inline bool operator!=(const ci_string& s1, const std::string& s2)
	{
		return s1 != ci_string(s2.begin(), s2.end());
	}
	inline bool operator!=(const std::string& s1, const ci_string& s2)
	{
		return ci_string(s1.begin(), s1.end()) != s2;
	}
	inline bool operator!=(const ci_wstring& s1, const std::wstring& s2)
	{
		return s1 != ci_wstring(s2.begin(), s2.end());
	}
	inline bool operator!=(const std::wstring& s1, const ci_wstring& s2)
	{
		return ci_wstring(s1.begin(), s1.end()) != s2;
	}
	inline bool operator<(const ci_string& s1, const std::string& s2)
	{
		return s1 < ci_string(s2.begin(), s2.end());
	}
	inline bool operator<(const std::string& s1, const ci_string& s2)
	{
		return ci_string(s1.begin(), s1.end()) < s2;
	}
	inline bool operator<(const ci_wstring& s1, const std::wstring& s2)
	{
		return s1 < ci_wstring(s2.begin(), s2.end());
	}
	inline bool operator<(const std::wstring& s1, const ci_wstring& s2)
	{
		return ci_wstring(s1.begin(), s1.end()) < s2;
	}
}

using lib::operator==;
using lib::operator!=;
using lib::operator<;

#endif // LIB_STRING