// xml.h - v1.0.2
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_XML
#define LIB_XML

#include <cstddef>
#include <istream>
#include <ostream>
#include <list>
#include <vector>
#include <utility>
#include <string>

namespace lib 
{
	template <typename CharT>
	class xml_element
	{
	public:
		// types
		typedef std::basic_string<CharT> string;
		typedef std::list<xml_element*> element_list;
		typedef std::pair<string, string> attribute;
		typedef std::list<attribute> attribute_list;

	public:
		// construction
		xml_element() {}
		
	public:
		// operations
		const string& name() const { return iName; }
		const attribute_list& attributes() const { return iAttributes; }
		const string& content() const { return iContent; }
		const element_list& children() const { return iChildren; }
		const string& attribute_value(const string& aAttributeName) const;
		string& name() { return iName; }
		attribute_list& attributes() { return iAttributes; }
		string& content() { return iContent; }
		element_list& children() { return iChildren; }
		string& attribute_value(const string& aAttributeName);

	private:
		// attributes
		string iName;
		attribute_list iAttributes;
		string iContent;
		element_list iChildren;
	};

	template <typename CharT>
	class basic_xml
	{
	public:
		// types
		typedef xml_element<CharT> element;
		typedef std::basic_string<CharT> string;
		typedef std::list<element*> element_list;
		typedef std::pair<string, string> attribute;
		typedef std::list<attribute> attribute_list;
		typedef std::pair<string, string> entity;
		typedef std::vector<entity> entity_list;

	public:
		// construction
		basic_xml();

	public:
		// operations
		void clear();
		const element& root() const { return iRoot; }
		element& root() { return iRoot; }
		element& insert(element& aParent, typename element_list::iterator aPosition, const CharT* aName);
		element& append(element& aParent, const CharT* aName);
		void erase(element& aParent, typename element_list::iterator aPosition);
		typename element_list::iterator find(element& aParent, const CharT* aName);
		bool read(std::basic_istream<CharT>& aStream);
		bool write(std::basic_ostream<CharT>& aStream);
		bool error() const { return iError; }
		void set_indent(CharT aIndentChar, std::size_t aIndentCount = 1);

	private:
		// implementation
		typedef std::pair<typename string::const_iterator, typename string::const_iterator> tag;
		typedef std::pair<typename string::const_iterator, typename string::const_iterator> content;
		tag next_tag(typename string::const_iterator aNext, typename string::const_iterator aDocumentEnd);
		typename string::const_iterator parse(element& aElement, const tag& aStartTag, typename string::const_iterator aDocumentEnd);
		void write_element(std::basic_ostream<CharT>& aStream, element& aElement, std::size_t aIndent);
		string parse_entities(const string& aString);
		string generate_entities(const string& aString);
	private:
		// attributes
		bool iError;
		element iRoot;
		typedef std::list<element> element_pool;
		element_pool iPool;
		entity_list iEntities;
		CharT iIndentChar;
		std::size_t iIndentCount;

		enum { PredefinedEntityCount = 5 };
		struct predefined_entity { const CharT* iFirst; const CharT* iSecond; };
		static const predefined_entity sPredefinedEntities[PredefinedEntityCount];
		static const string sTabChar;
		static const string sSpaceChar;
		static const string sLessThanChar;
		static const string sGreaterThanChar;
		static const string sEqualsChar;
		static const string sForwardSlashChar;
		static const string sAmpersandChar;
		static const string sSemicolonChar;
		static const string sHashChar;
		static const string sHexChar;
		static const string sQuoteChar;
		static const string sNameDelimeter;
		static const string sNameBadDelimeter;
		static const string sAttributeNameDelimeter;
		static const string sAttributeValueDelimeter;
		static const string sTagDelimeter;
		static const string sElementTagStart;
		static const string sElementTagEnd;
		static const string sWhiteSpace;
		static const string sCommentStart;
		static const string sCommentEnd;
		static const string sDeclarationStart;
		static const string sDeclarationEnd;
		static const string sEmptyTagWithAttributes;
		static const string sEmptyTag;
	};

	typedef basic_xml<char> xml;
	typedef basic_xml<wchar_t> wxml;
}

#include "xml_bits.h"

#endif // LIB_XML
