// string.h - v1.0.8
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_STRING
#define LIB_STRING

#include <algorithm>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cwchar>
#include <cctype>
#include <cwctype>

namespace lib 
{
	template <typename C, typename FwdIter1, typename FwdIter2>
	inline FwdIter1 tokens(FwdIter1 aFirst, FwdIter1 aLast, FwdIter2 aDelimeterFirst, FwdIter2 aDelimiterLast, C& aTokens, size_t aMaxTokens = 0, bool aMergeDelimeters = true)
	{
		if (aFirst >= aLast)
			return aFirst;

		typedef typename C::value_type value_type;

		FwdIter1 b = aFirst;
		FwdIter1 e = std::find_first_of(b, aLast, aDelimeterFirst, aDelimiterLast);
		size_t tokens = 0;
		while(e != aLast && (aMaxTokens == 0 || tokens < aMaxTokens))
		{
			if (b == e && !aMergeDelimeters)
			{
				aTokens.push_back(value_type(b, b));
				++tokens;
			}
			else if (b != e)
			{
				aTokens.push_back(value_type(b, e));
				++tokens;
			}
			b = e;
			++b;
			e = std::find_first_of(b, aLast, aDelimeterFirst, aDelimiterLast);
		}
		if (b != e && (aMaxTokens == 0 || tokens < aMaxTokens))
		{
			aTokens.push_back(value_type(b, e));
			b = e;
		}
		return b;
	}
	template <typename CharT, typename C>
	inline void tokens(const std::basic_string<CharT>& aLine, const std::basic_string<CharT>& aDelimeter, C& aTokens, size_t aMaxTokens = 0, bool aMergeDelimeters = true)
	{
		tokens(aLine.begin(), aLine.end(), aDelimeter.begin(), aDelimeter.end(), aTokens, aMaxTokens, aMergeDelimeters);
	}

	template <typename CharT>
	inline void remove_trailing(std::basic_string<CharT>& aString, const std::basic_string<CharT>& aTrailing)
	{
		typename std::basic_string<CharT>::size_type pos = aString.find_last_not_of(aTrailing);
		if (pos != std::basic_string<CharT>::npos)
			aString.erase(aString.begin() + pos + 1, aString.end());
	}

	template <typename CharT>
	inline long string_to_integer(const std::basic_string<CharT>& aString, size_t aBase = 10)
	{
		return strtol(aString.c_str(), 0, aBase);
	}

	template <>
	inline long string_to_integer<wchar_t>(const std::basic_string<wchar_t>& aString, size_t aBase)
	{
		return wcstol(aString.c_str(), 0, aBase);
	}

	template <typename CharT>
	inline std::basic_string<CharT> integer_to_string(long aInteger, size_t aBase = 10, size_t aWidth = 0, CharT aFill = '0')
	{
		std::basic_stringstream<CharT> string;
		if (aBase == 16)
			string << std::hex << std::uppercase;
		if (aWidth != 0)
		{
			string.width(aWidth);
			string.fill(aFill);
		}
		string << aInteger;
		return string.str();
	}

	template <>
	inline std::basic_string<wchar_t> integer_to_string<wchar_t>(long aInteger, size_t aBase, size_t aWidth, wchar_t aFill)
	{
		std::basic_stringstream<wchar_t> string;
		if (aBase == 16)
			string << std::hex;
		if (aWidth != 0)
		{
			string.width(aWidth);
			string.fill(aFill);
		}
		string << aInteger;
		return string.str();
	}

	template <typename CharT>
	inline unsigned long string_to_unsigned_integer(const std::basic_string<CharT>& aString, size_t aBase = 10)
	{
		return strtoul(aString.c_str(), 0, aBase);
	}

	template <>
	inline unsigned long string_to_unsigned_integer<wchar_t>(const std::basic_string<wchar_t>& aString, size_t aBase)
	{
		return wcstoul(aString.c_str(), 0, aBase);
	}

	template <typename CharT>
	inline std::basic_string<CharT> unsigned_integer_to_string(unsigned long aInteger, size_t aBase = 10, size_t aWidth = 0, CharT aFill = '0')
	{
		std::basic_stringstream<CharT> string;
		if (aBase == 16)
			string << std::hex << std::uppercase;
		if (aWidth != 0)
		{
			string.width(aWidth);
			string.fill(aFill);
		}
		string << aInteger;
		return string.str();
	}

	template <>
	inline std::basic_string<wchar_t> unsigned_integer_to_string<wchar_t>(unsigned long aInteger, size_t aBase, size_t aWidth, wchar_t aFill)
	{
		std::basic_stringstream<wchar_t> string;
		if (aBase == 16)
			string << std::hex;
		if (aWidth != 0)
		{
			string.width(aWidth);
			string.fill(aFill);
		}
		string << aInteger;
		return string.str();
	}

	inline std::string parse_escapes(const std::string& aString)
	{
		std::string ret = aString;
		std::string::size_type escapePos;
		while((escapePos = ret.find("\\r")) != std::string::npos)
			ret.replace(escapePos, 2, "\r");
		while((escapePos = ret.find("\\n")) != std::string::npos)
			ret.replace(escapePos, 2, "\n");
		return ret;
	}

	typedef unsigned long unicode_char_t;

	namespace utf16
	{
		inline bool is_high_surrogate(wchar_t aCharacter)
		{
			return aCharacter >= 0xD800 && aCharacter <= 0xDBFF;
		}
		inline bool is_low_surrogate(wchar_t aCharacter)
		{
			return aCharacter >= 0xDC00 && aCharacter <= 0xDFFF;
		}
		inline bool is_surrogate_pair(wchar_t aHighValue, wchar_t aLowValue)
		{
			return is_high_surrogate(aHighValue) && is_low_surrogate(aLowValue);
		}
	}

	inline std::string wide_to_utf8(const std::wstring& aString)
	{
		std::string narrowString;
		for (std::wstring::const_iterator i = aString.begin(); i != aString.end();)
		{
			unicode_char_t uch = *i++;
			if (utf16::is_high_surrogate(uch) && i != aString.end() && utf16::is_surrogate_pair(uch, *i))
			{
				uch = ((uch & 0x3FF) << 10);
				uch = uch | (*i++ & 0x3FF);
				uch += 0x10000;
			}
			if (uch <= 0x7F)
				narrowString.append(1, static_cast<char>(uch));
			else if (uch <= 0x7FF)
			{
				narrowString.append(1, static_cast<char>(((uch >> 6) & 0x1F) | 0xC0));
				narrowString.append(1, static_cast<char>((uch & 0x3F) | 0x80));
			}
			else if (uch <= 0xFFFF)
			{
				narrowString.append(1, static_cast<char>(((uch >> 12) & 0x0F) | 0xE0));
				narrowString.append(1, static_cast<char>(((uch >> 6 ) & 0x3F) | 0x80));
				narrowString.append(1, static_cast<char>((uch& 0x3F) | 0x80));
			}
			else
			{
				narrowString.append(1, static_cast<char>(((uch >> 18) & 0x07) | 0xF0));
				narrowString.append(1, static_cast<char>(((uch >> 12 ) & 0x3F) | 0x80));
				narrowString.append(1, static_cast<char>(((uch >> 6 ) & 0x3F) | 0x80));
				narrowString.append(1, static_cast<char>((uch& 0x3F) | 0x80));
			}
		}
		return narrowString;
	}

	namespace detail
	{
		template <typename FwdIter>
		inline unicode_char_t next_wide_bits(unicode_char_t aUnicodeChar, size_t aCount, FwdIter& aCurrent, FwdIter aEnd)
		{
			unicode_char_t unicodeChar = aUnicodeChar;
			FwdIter start = aCurrent;
			for (size_t i = 0; i != aCount; ++i)
			{
				++aCurrent;
				if (aCurrent == aEnd)
				{
					aCurrent = start;
					return L'?';
				}
				unsigned char nch = static_cast<unsigned char>(*aCurrent);
				if ((nch & 0xC0) == 0x80)
					unicodeChar = (unicodeChar << 6) | (static_cast<unicode_char_t>(nch & ~0xC0));
				else
				{
					aCurrent = start;
					return L'?';
				}
			}
			return unicodeChar;
		}
	}
	inline std::wstring utf8_to_wide(const std::string& aString)
	{
		std::wstring wideString;
		for (std::string::const_iterator i = aString.begin(); i != aString.end();)
		{
			unsigned char nch = static_cast<unsigned char>(*i);
			unicode_char_t uch = 0;
			if ((nch & 0x80) == 0)
				uch = static_cast<unicode_char_t>(nch & 0x7F);
			else
			{
				std::string::const_iterator old = i;
				if ((nch & 0xE0) == 0xC0)
					uch = detail::next_wide_bits(static_cast<unicode_char_t>(nch & ~0xE0), 1, i, aString.end());
				else if ((nch & 0xF0) == 0xE0)
					uch = detail::next_wide_bits(static_cast<unicode_char_t>(nch & ~0xF0), 2, i, aString.end());
				else if ((nch & 0xF8) == 0xF0)
					uch = detail::next_wide_bits(static_cast<unicode_char_t>(nch & ~0xF8), 3, i, aString.end());
				else if ((nch & 0xFC) == 0xF8)
					uch = detail::next_wide_bits(static_cast<unicode_char_t>(nch & ~0xFC), 4, i, aString.end());
				else if ((nch & 0xFE) == 0xFC)
					uch = detail::next_wide_bits(static_cast<unicode_char_t>(nch & ~0xFE), 5, i, aString.end());
				else if ((nch & 0xFF) == 0xFE)
					uch = detail::next_wide_bits(static_cast<unicode_char_t>(nch & ~0xFF), 6, i, aString.end());
				if (i == old)
					uch = static_cast<unicode_char_t>(nch);
			}
		
			if (uch <= 0xFFFF)
				wideString.append(1, uch);
			else
			{
				// UTF-16 bit...
				uch -= 0x10000;
				wideString.append(1, 0xd800|(uch >> 10));
				wideString.append(1, 0xdc00|(uch & 0x3FF));
			}

			if (i != aString.end())
				++i;
		}
		return wideString;
	}

	inline bool is_utf8_trailing(char aCharacter)
	{
		return (aCharacter & 0xC0) == 0x80;
	}

	template <typename C>
	inline std::string wide_to_narrow(const C& aWideString)
	{
		std::string ret;
		for (typename C::const_iterator i = aWideString.begin(); i != aWideString.end(); ++i)
			ret += (*i <= 0xFF ? *i : '?');
		return ret;
	}

	template <typename C>
	inline std::wstring narrow_to_wide(const C& aNarrowString)
	{
		std::wstring ret;
		for (typename C::const_iterator i = aNarrowString.begin(); i != aNarrowString.end(); ++i)
			ret += static_cast<wchar_t>(*i);
		return ret;
	}

	template <typename CharT>
	inline CharT tolower(int c, CharT=0) {	return ::tolower(c); }
	template <>
	inline wchar_t tolower<wchar_t>(int c, wchar_t) { return ::towlower(c); }

	template <typename Traits>
	struct ci_char_traits : Traits
	{
	public:
		typedef typename Traits::char_type char_type;
		typedef typename Traits::int_type int_type;
	public:
		static int compare(const char_type* s1, const char_type* s2, size_t n)
		{
			for(;n-- > 0;++s1, ++s2)
			{
				if (eq(*s1, *s2))
					continue;
				else if (lt(*s1, *s2))
					return -1;
				else
					return 1;
			}
			return 0;
		}
		static const char_type* find(const char_type* str, size_t n, const char_type& c)
		{
			while(n > 0)
			{
				if (eq(*str, c))
					return str;
				++str;
				--n;	
			}
			return 0;
		}
		static bool eq(const char_type& c1, const char_type& c2)
		{
			return lower(c1) == lower(c2);
		}
		static bool lt(const char_type& c1, const char_type& c2)
		{
			return lower(c1) < lower(c2);
		}
		static int_type lower(char_type c)
		{
			return tolower<char_type>(to_int_type(c), char_type());
		}
	};

	typedef std::basic_string<char, ci_char_traits<std::char_traits<char> > > ci_string;
	typedef std::basic_string<wchar_t, ci_char_traits<std::char_traits<wchar_t> > > ci_wstring;

	inline ci_string make_ci_string(const std::string& s)
	{
		return ci_string(s.begin(), s.end());
	}
	inline ci_wstring make_ci_string(const std::wstring& s)
	{
		return ci_wstring(s.begin(), s.end());
	}

	inline bool operator==(const ci_string& s1, const std::string& s2)
	{
		return s1 == ci_string(s2.begin(), s2.end());
	}
	inline bool operator==(const std::string& s1, const ci_string& s2)
	{
		return ci_string(s1.begin(), s1.end()) == s2;
	}
	inline bool operator==(const ci_wstring& s1, const std::wstring& s2)
	{
		return s1 == ci_wstring(s2.begin(), s2.end());
	}
	inline bool operator==(const std::wstring& s1, const ci_wstring& s2)
	{
		return ci_wstring(s1.begin(), s1.end()) == s2;
	}
	inline bool operator!=(const ci_string& s1, const std::string& s2)
	{
		return s1 != ci_string(s2.begin(), s2.end());
	}
	inline bool operator!=(const std::string& s1, const ci_string& s2)
	{
		return ci_string(s1.begin(), s1.end()) != s2;
	}
	inline bool operator!=(const ci_wstring& s1, const std::wstring& s2)
	{
		return s1 != ci_wstring(s2.begin(), s2.end());
	}
	inline bool operator!=(const std::wstring& s1, const ci_wstring& s2)
	{
		return ci_wstring(s1.begin(), s1.end()) != s2;
	}
	inline bool operator<(const ci_string& s1, const std::string& s2)
	{
		return s1 < ci_string(s2.begin(), s2.end());
	}
	inline bool operator<(const std::string& s1, const ci_string& s2)
	{
		return ci_string(s1.begin(), s1.end()) < s2;
	}
	inline bool operator<(const ci_wstring& s1, const std::wstring& s2)
	{
		return s1 < ci_wstring(s2.begin(), s2.end());
	}
	inline bool operator<(const std::wstring& s1, const ci_wstring& s2)
	{
		return ci_wstring(s1.begin(), s1.end()) < s2;
	}
}

using lib::operator==;
using lib::operator!=;
using lib::operator<;

#endif // LIB_STRING