// treesegment.h
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_TREESEGMENT
#define LIB_TREESEGMENT

#include "utility.h"
#include "segmented_array.h"
#include "memory.h"

namespace lib 
{
	template <typename T, size_t N, typename A = allocator<T>, typename A2 = allocator<T> >
	class tree_segment
	{
	public:
		typedef T value_type;
		typedef A allocator_type;
		typedef A2 allocator2_type;
		typedef value_type* pointer;
		typedef const value_type* const_pointer;
		typedef value_type& reference;
		typedef const value_type& const_reference;
		struct node;
	public:
		typedef segmented_array<void*, N, A2> node_segment;
		typedef typename node_segment::iterator iterator;
		typedef typename node_segment::const_iterator const_iterator;
		struct node
		{
			value_type iValue;
			node_segment iChildren;
			private: node() {} public:
			node(const value_type& value, const A2& allocator) : iValue(value), iChildren(allocator) {}
			node(const node& rhs) : iValue(rhs.iValue), iChildren(rhs.iChildren) {}
			value_type& value() { return iValue; }
			node_segment& children() { return iChildren; }
			bool empty() const { return iChildren.empty(); }
		};
		typedef node* node_pointer;
		typedef const node* const_node_pointer;

	private:
		A iAllocator;
		A2 iAllocator2;
		node* iRoot;

	public:
		enum { node_size = sizeof(node) };

	public:
		// construction
		tree_segment() : iRoot(0) {}
		tree_segment(const A& allocator, const A2& allocator2) : iAllocator(allocator), iAllocator2(allocator2), iRoot(0) {}
		~tree_segment() {}

		// traversals
		node_pointer root() { if (iRoot == 0) new_root(0); return iRoot; }
		const_node_pointer root() const { return iRoot; }
		bool empty() const { return iRoot == 0 || iRoot->iChildren.size() == 0; }
		// modifiers
		void new_root(node_pointer node)
		{
			if (iRoot != 0)
				erase(iRoot);
			if (node == 0)
				node = buy_node(value_type());
			iRoot = node;
		}
		void push_back(node_pointer parent, const value_type& value) 
		{			
			if (parent == 0)
			{
				if (iRoot == 0)
					new_root(0);
				parent = iRoot;
			}
			node_pointer new_node = buy_node(value);
			parent->iChildren.push_back(new_node);
		}
		void push_front(node_pointer parent, const value_type& value) 
		/* WARNING: This method is not O(1) compexity */
		{			
			if (parent == 0)
			{
				if (iRoot == 0)
					iRoot = buy_node(value_type());
				parent = iRoot;
			}
			node_pointer new_node = buy_node(value);
			parent->iChildren.insert(parent->iChildren.begin(), new_node);
		} 
		void erase(node_pointer position, node_pointer parent = 0) 
		{ 
			if (position == 0)
				return;

			if (position == iRoot)
				iRoot = 0;

			iterator first = position->iChildren.begin();
			iterator last = position->iChildren.end();
			while(first != last)
			{
				erase(static_cast<node_pointer>(*first), 0);
				first++;
			}

			iAllocator.destruct(position);
			iAllocator.deallocate(reinterpret_cast<A::pointer>(position));
			if (parent != 0)
				parent->children().remove(position, false);
		}
		void erase_children(node_pointer position)
		{
			if (position == 0)
				return;
			iterator first = position->children().begin();
			iterator last = position->children().end();
			while(first != last)
			{
				erase(static_cast<node_pointer>(*first++));
			}
			position->children().clear();
		}
	private:
		// implementation
		node_pointer buy_node(const value_type& value)
		{
			node_pointer new_node  = reinterpret_cast<node_pointer>(iAllocator.allocate(1, sizeof(node)));
			iAllocator.construct(new_node, node(value, iAllocator2));
			return new_node;
		}
	};
}

#endif // LIB_TREESEGMENT