// vecarray.h - v1.1.4
/* 
 * Copyright (c) 2008 Leigh Johnston.  All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this
 * software and its documentation for any purpose is hereby
 * granted, provided that the above copyright notice
 * appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation.
 * The author makes no representations about the
 * suitability of this software for any purpose. It is provided
 * "as is" without express or implied warranty.
 */

#ifndef LIB_VECARRAY
#define LIB_VECARRAY

#include <cassert>
#include <cstddef>
#include <stdexcept>
#include "type.h"
#include "memory.h"
#include "algorithm.h"

#define VECARRAY_PREVENT_OVERFLOW

namespace lib
{
	#ifdef VECARRAY_PREVENT_OVERFLOW
	struct vecarray_overflow : public std::exception
	{
		const char* what() { return "vecarray overflow"; }
	};
	#endif

	template<typename T, size_t N>
	class vecarray
	{
	public:
		typedef T value_type;
		typedef T* pointer;
		typedef const T* const_pointer;
		typedef T& reference;
		typedef const T& const_reference;
		typedef pointer iterator;
		typedef const_pointer const_iterator;
		typedef size_t size_type;
		typedef ptrdiff_t difference_type;
	private:
		template <typename IsScalar>
		struct vecarray_scalar_trait_impl
		{ 
			typedef pointer pointer;
			typedef IsScalar type;
			typedef const T& parameter_type; 
			inline static void construct(pointer object, parameter_type value)
			{
				new (static_cast<void*>(object)) T(value);
			}
			inline static void destruct(pointer object)
			{
				object->~T();
			}
			inline static void destruct(pointer first, pointer last)
			{
				while(first != last)
				{
					first->~T();
					++first;
				}
			}
		};
		template <>
		struct vecarray_scalar_trait_impl<true_type>
		{ 
			typedef pointer pointer;
			typedef true_type type;
			typedef T parameter_type; 
			inline static void construct(pointer object, parameter_type value)
			{
				*object = value;
			}
			inline static void destruct(pointer object)
			{
			}
			inline static void destruct(pointer first, pointer last)
			{
			}
		};
		typedef vecarray_scalar_trait_impl<typename is_scalar<T>::type> scalar_trait;
	public:
		typedef typename scalar_trait::parameter_type parameter_type;
	
	private:
		char iData[sizeof(T) * N];
		size_type iSize;
	
	public:
		// construction
		vecarray() : iSize(0)
		{
		}
		vecarray(const vecarray& rhs) : iSize(0)
		{
			insert(begin(), rhs.begin(), rhs.end());
		}
		#if _MSC_VER != 1200 // VC++ 6.0
		template <typename T2, size_t N2>
		vecarray(const vecarray<T2, N2>& rhs) : iSize(0)
		{
			insert(begin(), rhs.begin(), rhs.end());
		}
		#endif
		vecarray(size_type n, parameter_type value) : iSize(0)
		{
			insert(begin(), n, value);
		}
		template<typename InputIterator>
		vecarray(InputIterator first, InputIterator last) : iSize(0)
		{
			insert(begin(), first, last);
		}
		~vecarray()
		{
			clear();
		}
		// traversals
		iterator begin() { return reinterpret_cast<pointer>(iData); }
		const_iterator begin() const { return reinterpret_cast<const_pointer>(iData); }
		iterator end() { return reinterpret_cast<pointer>(iData) + iSize; }
		const_iterator end() const { return reinterpret_cast<const_pointer>(iData) + iSize; }
		iterator locate(size_type position) { return reinterpret_cast<pointer>(iData) + position; }
		const_iterator locate(size_type position) const { return reinterpret_cast<const_pointer>(iData) + position; }
		bool empty() const { return iSize == 0; }
		bool full() const { return iSize == N; }
		size_type size() const { return iSize; }
		size_type available() const { return N - iSize; }
		size_type capacity() const { return N; }
		size_type max_size() const { return N; }
		size_type after(size_type position) const { return position < iSize ? iSize - position : 0; }
		// element access
		reference operator[](size_type n) { return *(begin() + n); }
		const_reference operator[](size_type n) const { return *(begin() + n); }
		reference at(size_type n) { if (n < iSize) return operator[](n); else { throw_out_of_range("vecarray::at"); assert(false); return *((T*)0); } }
		const_reference at(size_type n) const { if (n < iSize) return operator[](n); else { throw_out_of_range("vecarray::at"); assert(false); return *((T*)0); } }
		reference front() { return *begin(); }
		reference back() { return *(begin() + (iSize - 1)); }
		const_reference front() const { return *begin(); }
		const_reference back() const { return *(begin() + (iSize - 1)); }
		// modifiers
		vecarray& operator=(const vecarray& rhs)
		{
			if (&rhs != this)
				assign(rhs.begin(), rhs.end());
			return *this;
		}
		#if _MSC_VER != 1200 // VC++ 6.0
		template<typename T2, size_t N2>
		vecarray& operator=(const vecarray<T2, N2>& rhs)
		{
			if (static_cast<const void*>(&rhs) != static_cast<const void*>(this))
				assign(rhs.begin(), rhs.end());
			return *this;
		}
		#endif
		template <typename InputIterator>
		void assign(InputIterator first, InputIterator last)
		{
			clear();
			insert(begin(), first, last);
		}
		void assign(size_type n, parameter_type value)
		{
			clear();
			insert(begin(), n, value);
		}
		iterator insert(iterator position, parameter_type value)
		{
			insert(position, &value, &value+1);
			return position;
		}
		void insert(iterator position, size_type n, parameter_type value)
		{
			while(n > 0)
			{
				insert(position, value);
				++position;
				--n;
			}
		}
		template <typename InputIterator>
		void insert(iterator position, InputIterator first, InputIterator last)
		{
			insert_dispatch(position, first, last, typename is_integral<InputIterator>::type());
		}
		void insert(iterator position, iterator first, iterator last)
		{
			insert(position, static_cast<const_iterator>(first), static_cast<const_iterator>(last));
		}
		void insert(iterator position, const_iterator first, const_iterator last)
		{
			difference_type n = last - first;
			overflow_check(n);
			iterator theEnd = end();
			difference_type t = theEnd - position;
			if (t > 0)
			{
				if (t > n)
				{
					uninitialized_copy(theEnd - n, theEnd, theEnd);
					copy_backward(position, theEnd - n, theEnd);
					copy(first, last, position);
				}
				else
				{
					uninitialized_copy(theEnd - t, theEnd, theEnd + (n - t));
					uninitialized_copy(first + t, last, theEnd);
					copy(first, first + t, position);
				}
			}
			else
			{
				uninitialized_copy(first, last, theEnd);
			}
			iSize += n;
		}
		iterator erase(iterator position) 
		{ 
			iterator last = end(); --last;
			copy(position + 1, end(), position);
			scalar_trait::destruct(last);
			--iSize;
			return position;
		}
		iterator erase(iterator first, iterator last) 
		{ 
			if (last != end())
				copy(last, end(), first);
			scalar_trait::destruct(end() - (last - first), end());
			iSize-=(last - first);
			return first;
		}
		void clear()
		{
			erase(begin(), end());
		}
		void push_back(parameter_type value)
		{
			#if _MSC_VER != 1200 // VC++ 6.0
			push_back_dispatch(value, typename scalar_trait::type());
			#else
			push_back_dispatch(value, scalar_trait::type());
			#endif
		}
		void pop_back()
		{
			erase(end() - 1);
		}
		void remove(parameter_type value, bool multiple = true)
		{
			for (iterator i = begin(); i != end(); )
			{
				if (*i == value)
				{
					erase(i);
					if (!multiple)
						return;
				}
				else
					++i;
			}
		}
		void resize(size_type n, parameter_type value = value_type())
		{
			while(iSize < n)
				push_back(value);
		}
		template <typename T2, size_t N2>
		void swap(vecarray<T2, N2>& rhs)
		{
			vecarray tmp = rhs;
			rhs = *this;
			*this = tmp;
		}
	
	private:
		template <typename InputIterator>
		void insert_dispatch(iterator position, InputIterator first, InputIterator last, false_type)
		{
			while(first != last)
				insert(position++, *first++);
		}
		template <typename Integer>
		void insert_dispatch(iterator position, Integer n1, Integer n2, true_type)
		{
			size_type n = static_cast<size_type>(n1);
			value_type value = static_cast<value_type>(n2);
			while(n > 0)
			{
				insert(position, value);
				++position;
				--n;
			}
		}
		void push_back_dispatch(parameter_type value, false_type)
		{
			overflow_check(1);
			scalar_trait::construct(end(), value);
			++iSize;
		}
		void push_back_dispatch(parameter_type value, true_type)
		{
			overflow_check(1);
			*end() = value;
			++iSize;
		}
		void throw_out_of_range(const char* s) const
		{
			throw std::out_of_range(s);
		}
		void overflow_check(size_type n) const
		{
			#ifdef VECARRAY_PREVENT_OVERFLOW
			if (iSize + n > N)
				throw vecarray_overflow();
			#endif
		}
	};
}

#endif // LIB_VECARRAY